
# Copyright (c) 2009 Greg Pintilie - pintilie@mit.edu

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import chimera
import chimera.match
import os
import os.path
import Tkinter
from CGLtk import Hybrid
import VolumeData
import _multiscale
import _surface
import numpy
import _contour
import Matrix
import FitMap
import VolumeViewer
from sys import stderr
from time import clock
from random import random as rand

from axes import prAxes, AxesMod
import regions

from segment_dialog import current_segmentation, segmentation_map
from Segger import getMod, NumberFromName, dev_menus, timing

OML = chimera.openModels.list

SAF_DVOL = 0.75
SAF_DBRAD = 0.3
SAF_LS_DEPTH = 4
SAF_LS_NGROUPS = 1000

REG_OPACITY = 0.45
MAX_NUM_GROUPS = 1000



def umsg ( txt ) :
    print txt
    status ( txt )

def status ( txt ) :
    txt = txt.rstrip('\n')
    msg.configure(text = txt)
    msg.update_idletasks()

class Fit_Segments_Dialog ( chimera.baseDialog.ModelessDialog ):

    title = "Fit To Segments (Segger v1.6)"
    name = "fit segments"
    buttons = ('Fit', 'Options', "Close")
    help = 'http://ncmi.bcm.edu/ncmi/software/segger/'

    def fillInUI(self, parent):

        tw = parent.winfo_toplevel()
        self.toplevel_widget = tw
        tw.withdraw()

        parent.columnconfigure(0, weight = 1)

        row = 1

        menubar = Tkinter.Menu(parent, type = 'menubar', tearoff = False)
        tw.config(menu = menubar)

        file_menu_entries = (
            ("Save fitted structure", self.SaveStrucFit),
            )

        fmenu = Hybrid.cascade_menu(menubar, 'File', file_menu_entries)

        self.UseAllMods = Tkinter.IntVar()
        self.UseAllMods.set ( 0 )

        self.strucCopyAfterFit = Tkinter.IntVar()
        self.strucCopyAfterFit.set ( 0 )

        fit_menu_entries = (
            #("Selected region", self.FitToSelRegion),
            #("Each selected region", self.FitToSelectedRegions),
            #("Combined selected regions", self.FitMapToSelRGroup),
            #("Regions around selected", self.FitMapToRegionsAroundSel),
            #("Groups of regions", self.FitMapToRGroups),
            #'separator',
            ("Show molecule axes", self.StrucShowAxes),
            ("Hide molecule axes", self.StrucHideAxes),
            #("Place best fits", self.PlaceBestFits),
            ("Show overlapping regions", self.ShowOverlappingRegions),
            )

        fmenu = Hybrid.cascade_menu(menubar, 'Fit', fit_menu_entries)

        if dev_menus:
            fmenu.add_separator()
            for lbl, var, val in (
                #("By principal axes", self.rotaSearch, 0),
                #("By rotation", self.rotaSearch, 1),
                ):
                fmenu.add_radiobutton(label = lbl, variable = var, value = val)
            for lbl, var in (("Chains = models", self.UseAllMods),
                             #("Save all fits", self.strucCopyAfterFit),
                             ):
                fmenu.add_checkbutton(label = lbl, variable = var)
            for lbl, cmd in (
                #("Replicate with Bio-Matrices", self.StrucBioMT),
                ("Sim chain maps", self.SimChainMaps),
                ("Make chain maps", self.StrucChainMaps),
                (" - show chain maps", self.StrucShowChainMaps),
                (" - hide chain maps", self.StrucHideChainMaps),
                (" - close chain maps", self.StrucCloseChainMaps),
                (" - overlap regions", self.StrucChMapsOvRegs),
                (" - close 1 chain map", self.StrucCloseChainMap),
                #(" - delete chain maps", self.StrucDelChainMaps),
                #(" - show chain/regions", self.ShowChRegs),
                ("Segmentation accuracy", self.SegAccuracy),
                ("Fit RMSD", self.FitRMSD),
                #("Sel Chains", self.GetSelectedChains),
                ("Align to selected chain", self.AlignToSel),
                ("Extract proteins", self.ExtractProteins),
                #("Shape match with selected regions", self.SelRegsShapeScore),
                #("Adjust threshold for best match", self.SelRegsOptimizeShapeScore),
                #("Best shape score", self.StrucBestShapeScore),
                ("Group regions", self.StrucGroupRegions),
                ("Next group", self.NextRGroup),
                #("Find group w/selected", self.FindGroupFromSelRegs),
                #("Get fits", self.GetFits),
                #("Next fit", self.NextFit),
                #("Best fit around selected", self.FitMapToRSelGroups),
                #("Fit ALL around selected", self.FitMapsToRegionsAroundSel),
                #("Fit ALL structures", self.FitMapsToRGroups),
                #("Displayed volume", self.StrucMapVolume),
                ("Fit to map (local)", self.FitSMapToDMap),
                ("Zero density map", self.ZeroDMap_with_FMap),
                ("Extract density", self.TakeDMap_with_FMap),
                ("Masked map", self.MaskedMap),
                ("Next bio Matrix", self.NextBioMt),
                ("All bio matrices", self.GoBioMt),
                             ):
                fmenu.add_command(label = lbl, command = cmd)

        from chimera.tkgui import aquaMenuBar
        aquaMenuBar(menubar, parent, row = 0, columnspan=3)


        f = Tkinter.Frame(parent)
        f.grid(column=0, row=row, sticky='ew')
        row += 1

        l = Tkinter.Label(f, text='Structure to fit')
        l.grid(column=0, row=0, sticky='w')
 
        self.struc = Tkinter.StringVar(parent)
        self.strucMB  = Tkinter.Menubutton ( f, textvariable=self.struc, relief=Tkinter.RAISED )
        self.strucMB.grid (column=1, row=0, sticky='we', padx=5)
        self.strucMB.menu  =  Tkinter.Menu ( self.strucMB, tearoff=0, postcommand=self.StrucMenu )
        self.strucMB["menu"]  =  self.strucMB.menu

        row += 1

        self.struc_map = Tkinter.StringVar(parent)


        op = Hybrid.Popup_Panel(parent)
        opf = op.frame
        opf.grid(row = row, column = 0, sticky = 'news')
        opf.grid_remove()
        opf.columnconfigure(0, weight=1)
        self.optionsPanel = op.panel_shown_variable
        row += 1
        orow = 0

        cb = op.make_close_button(opf)
        cb.grid(row = orow, column = 0, sticky = 'e')

        l = Tkinter.Label(opf, text='Fitting Options', font = 'TkCaptionFont')
        l.grid(column=0, row=orow, sticky='w', pady=5)
        orow += 1

        fopt = Tkinter.Frame(opf)
        fopt.grid(column=0, row=orow, sticky='ew', padx=10)
        orow += 1
        forow = 0

        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='e')
        forow += 1

        l = Tkinter.Label(f, text='Density map resolution:')
        l.grid(column=0, row=0, sticky='w')

        self.simRes = Tkinter.StringVar(fopt)
        e = Tkinter.Entry(f, width=4, textvariable=self.simRes)
        e.grid(column=1, row=0, sticky='w', padx=5)

        l = Tkinter.Label(f, text='grid spacing:')
        l.grid(column=2, row=0, sticky='w')

        self.simGridSp = Tkinter.StringVar(fopt)
        e = Tkinter.Entry(f, width=4, textvariable=self.simGridSp)
        e.grid(column=3, row=0, sticky='w', padx=5)

        b = Tkinter.Button(f, text="Calculate Map", command=self.GenStrucMap)
        b.grid (column=4, row=0, sticky='w', padx=5)



        forow += 1

        l = Tkinter.Label(fopt, text='Which regions to align the structure to:')
        l.grid(column=0, row=forow, sticky='w')

        forow += 1

        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='w')

        self.alignTo = Tkinter.StringVar()
        self.alignTo.set ( 'combined_selected_regions' )

        l = Tkinter.Label(f, text=' ', width=5)
        l.grid(column=0, row=0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Combined selected regions", variable=self.alignTo, value = 'combined_selected_regions')
        c.grid (column=1, row=0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Each selected region", variable=self.alignTo, value = 'each_selected_region')
        c.grid (column=1, row=1, sticky='w')

        c = Tkinter.Radiobutton(f, text="Groups of regions including selected region", variable=self.alignTo, value = 'around_selected')
        c.grid (column=1, row=2, sticky='w')

        c = Tkinter.Radiobutton(f, text="Groups of regions including all regions", variable=self.alignTo, value = 'all_groups')
        c.grid (column=1, row=3, sticky='w')



        forow += 1

        l = Tkinter.Label(fopt, text='Alignment Method:')
        l.grid(column=0, row=forow, sticky='w')

        forow += 1

        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='w')

        self.rotaSearch = Tkinter.IntVar()
        self.rotaSearch.set ( 0 )

        l = Tkinter.Label(f, text=' ', width=5)
        l.grid(column=0, row=0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Align principal axes (faster)", variable=self.rotaSearch, value = 0)
        c.grid (column=1, row = 0, sticky='w')

        c = Tkinter.Radiobutton(f, text="Rotational search (more reliable)", variable=self.rotaSearch, value = 1)
        c.grid (column=1, row = 1, sticky='w')



        forow += 1

        c = Tkinter.Checkbutton(fopt, text="Save each fit when doing multiple alignments",
                                variable=self.strucCopyAfterFit)
        c.grid (row = forow, column=0, sticky='w')

        forow += 1


        f = Tkinter.Frame(fopt)
        f.grid(column=0, row=forow, sticky='w')
        forow += 1

        l = Tkinter.Label( f, text="Top number of fits to place" ) 
        l.grid(column=0, row=0, sticky='e')

        self.numFits = Tkinter.StringVar(opf)
        self.numFits.set ( '1' )
        e = Tkinter.Entry(f, width=3, textvariable=self.numFits)
        e.grid(column=1, row=0, sticky='w', padx=5)

        b = Tkinter.Button(f, text="Place", command=self.PlaceBestFits)
        b.grid (column=2, row=0, sticky='w', padx=5)


        dummyFrame = Tkinter.Frame(parent, relief='groove', borderwidth=1)
        Tkinter.Frame(dummyFrame).pack()
        dummyFrame.grid(row=row,column=0,columnspan=7, pady=7, sticky='we')

        row = row + 1

        global msg
        msg = Tkinter.Label(parent, width = 60, anchor = 'w', justify = 'left')
        msg.grid(column=0, row=row, sticky='ew')
        self.msg = msg
        row += 1

        self.SetResolution()

        for m in regions.segmentations() :

            try :
                self.struc.set ( m.mol.name )
                self.struc_map.set ( m.name )
            except : pass

            try : m.seg_map
            except : continue

        chimera.openModels.addRemoveHandler(self.ModelClosed, None)


    def Options(self):

        self.optionsPanel.set(not self.optionsPanel.get())


    def ModelClosed(self, trigger, n, mlist):

        # Clear menus that are showing closed models.
        for mvar in (self.struc, self.struc_map):
            if len( mvar.get() ) > 0 and getMod ( mvar.get() ) is None:
                mvar.set('')


    def SetResolution ( self ):

        dmap = segmentation_map()
        if dmap == None : return

        if len ( self.simRes.get() ) == 0:
            res = min(dmap.data.step) * 3
            self.simRes.set ( '%.3g' % res )
            self.simGridSp.set ( '%.3g' % (res/3.0) )


    def CurrentSegmentation ( self, warn = True ):

        return current_segmentation(warn)

        
    def StrucMenu ( self ) :

        self.strucMB.menu.delete ( 0, 'end' )   # Clear menu

        self.strucMB.menu.add_radiobutton ( label="Open structures:" )
        self.strucMB.menu.add_separator()

        slist = []
        open_mols = {}
        for m in OML() :
            if type(m) == chimera.Molecule and m.name.find("_graph") == -1:
                fn = NumberFromName ( m.name, 'f' )
                slist.append ( [fn, m.name] )
                open_mols[m.name] = True

        slist.sort ()
        cur_sel_found = False
        for fn, mname in slist :
            self.strucMB.menu.add_radiobutton ( label=mname, variable=self.struc,
                                                command = self.StrucSelected )

            if mname == self.struc.get() :
                cur_sel_found = True

        if cur_sel_found == False :
            self.struc.set ( "" )

        dmap = segmentation_map()
        if dmap == None : return

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        files = os.listdir ( path );
        mols_in_path = []
        for f in files :
            if f.find ( ".pdb" ) >= 0 and open_mols.has_key(f) == False :
                mols_in_path.append ( f )

        if len ( mols_in_path ) == 0 : return

        self.strucMB.menu.add_separator()
        self.strucMB.menu.add_radiobutton ( label="In %s:" % path )
        self.strucMB.menu.add_separator()

        for fm in mols_in_path :
            self.strucMB.menu.add_radiobutton ( label=fm, variable=self.struc,
                                                command = self.StrucSelected )


    def StrucSelected ( self ) :
        print "selected structure", self.struc.get()
        fname = self.struc.get()

        fmol = getMod ( fname )
        if fmol :
            print " - open"

            for om in chimera.openModels.list() :
                try : om_mol = om.mol
                except : continue
                if om_mol == fmol :
                    self.struc_map.set ( om.name )
                    break

            return

        dmap = segmentation_map()
        if dmap == None : print " - no map selected"; return

        path = os.path.dirname(dmap.data.path) + os.path.sep
        
        fmol = chimera.openModels.open ( path + fname )[0]
        print " - opened"
        fmol.ch_colors = RandColorChains ( fmol )



    def StrucMapSelected ( self ) :

        print "struc map selected", self.struc_map.get()
        res = NumberFromName ( self.struc_map.get(), 'r' )
        if res != None :
            self.simRes.set ( "%.1f" % res )
            sp = NumberFromName ( self.struc_map.get(), 'sp' )
            if sp != None :
                self.simRes.set ( "%.0f" % (res,) )
                self.simGridSp.set ( "%.1f" % (sp,) )

        mm = getMod ( self.struc_map.get() )
        mm.display = True

        try : mm.mol
        except : return

        mm.mol.display = True

        self.struc.set ( mm.mol.name )

        for m in OML () :

            if type(m) == chimera.Molecule : continue

            if type(m) == VolumeViewer.volume.Volume and m != mm :
                m.display = False

            try : m.mol
            except : continue
            if m != mm :
                m.display = False
                m.mol.display = False



    def Fit ( self ) :

        alignTo = self.alignTo.get()

        methods = {
         'combined_selected_regions' :
           ("Fitting to combined selected regions...", self.FitMapToSelRGroup),
         'each_selected_region' : 
           ("Fitting to each selected region...", self.FitToSelectedRegions),
         'around_selected' :
           ("Fitting groups around selected...", self.FitMapToRegionsAroundSel),
         'all_groups' :
             ( "Fitting to all groups...", self.FitMapToRGroups)
         }

        if alignTo in methods:
             descrip, func = methods[alignTo]
             print descrip
             func()
             dmap = segmentation_map()
             if dmap:
                 dmap.display = False


    def StrucCenter ( self ) :

        if len(self.struc.get()) == 0 : umsg ("Please select a structure first"); return
        fmol = getMod ( self.struc.get() )
        if fmol == None : umsg ( "%s not open" % self.struc.get() ); return

        centerMol ( fmol )

        mm = segmentation_map()
        if mm : fmol.openState.xform = mm.openState.xform



    def StrucShowAxes ( self ) :

        if len ( self.struc.get() ) == 0 :
            print "Please select a structure"
            return
        
        print "Showing axes for", self.struc.get()

        fmol = getMod ( self.struc.get() )
        if fmol == None :
            print self.struc.get(), "not open";
            return

        try :
            chimera.openModels.close ( fmol.axes )
            fmol.axes = None
        except :
            pass

        if not hasattr(fmol, 'centered'):
            centerMol(fmol)
        print fmol.COM
        print fmol.Extents

        #fmol.axes = AxesMod ( fmol.COM, fmol.U, fmol.Extents, 1.0 )
        fmol.axes =  AxesMod ( Extents = fmol.Extents, rad = 1.0,
                               alignTo = fmol )
        fmol.axes.name = os.path.splitext (fmol.name)[0] + "_axes"



    def StrucHideAxes ( self ) :

        fmol = getMod ( self.struc.get() )
        if fmol == None : print self.struc.get(), "not open"; return

        try :
            chimera.openModels.close ( fmol.axes )
            fmol.axes = None
        except :
            pass


    def SelStrucMap ( self, sel_str, mol, dmap, color ) :

        atoms = chimera.selection.OSLSelection( sel_str ).atoms()

        if len(atoms) == 0 :
            print "- empty"
            return None
        
        points = _multiscale.get_atom_coordinates ( atoms, transformed = True )
        _contour.affine_transform_vertices ( points, Matrix.xform_matrix(dmap.openState.xform.inverse()) )

        step = chimera.Vector ( dmap.data.step[0], dmap.data.step[1], dmap.data.step[2] )
        sg = VolumeData.zone_masked_grid_data ( dmap.data, points, 2.0 )

        sgm = sg.matrix()
        #sgmt = numpy.where ( sgm > 0, sgm, numpy.ones_like(sgm)*1e99 )
        #min_d = numpy.min ( sgmt )

        #sgmt = numpy.where ( sgm > 0, numpy.ones_like(sgm), numpy.zeros_like(sgm) )
        #sgd = VolumeData.Array_Grid_Data (sgmt, dmap.data.origin, dmap.data.step, dmap.data.cell_angles, dmap.data.rotation)

        try : gv = VolumeViewer.volume.add_data_set ( sg, None )
        except : gv = VolumeViewer.volume.volume_from_grid_data ( sg )
        gv.openState.xform = dmap.openState.xform

        dvals = dmap.interpolated_values ( points, mol.openState.xform )
        min_d = numpy.min ( dvals ) / 2.0
        min_d = dmap.surface_levels[0]

        print " - min d: %.4f," % min_d,

        gv.region = ( gv.region[0], gv.region[1], [1,1,1] )

        gv.display = True
        gv.surface_levels = [ min_d ]
        #gv.surface_levels = [20.0]
        gv.surface_colors = [ color ]
        gv.surf_color = color
        ro = VolumeViewer.volume.Rendering_Options()
        ro.surface_smoothing = True
        #ro.smoothing_factor = .25
        #ro.smoothing_iterations = 2
        gv.update_surface ( False, ro )
        gv.atoms = atoms

        for sp in gv.surfacePieces :
            v, t = sp.geometry
            if len(v) == 8 and len(t) == 12 : sp.display = False

        cm = gv.data.matrix()

        nze = numpy.nonzero ( cm )
        print "%d NZ \\" % len ( nze[0] ),

        cmt = numpy.where ( cm > min_d, cm, numpy.zeros_like(cm) )
        nze = numpy.nonzero ( cmt )
        print " %d NZ" % len ( nze[0] )
        gv.d_thr = min_d

        gv.imap = {}
        for ei, i in enumerate ( nze[0] ) :
            j = nze[1][ei]
            k = nze[2][ei]

            try : mi = gv.imap[i]
            except : mi = {}; gv.imap[i] = mi

            try : mij = mi[j]
            except : mij = {}; mi[j] = mij

            mij[k] = 1


        com = numpy.sum(points, axis=0) / len(points)
        gv.COM = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points_v = points - comv
        gv.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

        return gv



    def MakeChainMaps ( self, mols, dmap ) :

        try : print len(dmap.chain_maps), "chain maps so far"
        except : dmap.chain_maps = []


        if len(mols) == 1 :

            mol = mols[0]
            print "Making chain maps for %s in %s:" % (mol.name, dmap.name)

            for cid, clr in mol.chain_colors.iteritems() :

                map_name = os.path.splitext ( dmap.name )[0]
                cname = map_name + "_" + cid
                sel_str = "#%d:.%s" % (mol.id, cid)
                print "%s [%s]" % (cname, sel_str),

                gv = self.SelStrucMap ( sel_str, mol, dmap, clr.rgba() )
                if gv :
                    gv.name = cname
                    gv.chain_id = cid
                    dmap.chain_maps.append ( gv )
                    gv.mol = mol

                #return


        else :

            print "Making struc maps for %d strucs" % len(mols)

            for mol in mols :

                map_name = os.path.splitext ( dmap.name )[0]
                mol_name = os.path.splitext ( mol.name )[0]
                cname = map_name + "_" + mol_name
                sel_str = "#%d" % (mol.id)
                print "%s [%s]" % (cname, sel_str),
                
                clr = (rand(), rand(), rand(), 1.0)
                gv = self.SelStrucMap ( sel_str, mol, dmap, clr )
                gv.name = cname
                gv.chain_id = mol_name
                dmap.chain_maps.append ( gv )




    def StrucChainMaps ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        mols = []
        if self.UseAllMods.get() :
            for mod in chimera.openModels.list() :
                if type(mod) == chimera.Molecule :
                    mols.append ( mod )

        else :
            m = getMod ( self.struc.get() )
            if m != None :
                mols.append ( m )
            m.chain_colors = RandColorChains ( m )

        if len(mols) == 0 :
            print "No structures"; return

        self.MakeChainMaps ( mols, dmap )
        print "- %d chain or unit maps" % len ( dmap.chain_maps )



    def StrucShowChainMaps ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        print "%s - showing %d chain maps" % (dmap.name, len(dmap.chain_maps))
        for chm in dmap.chain_maps :
            chm.display = True


    def StrucHideChainMaps ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        print "%s - hiding %d chain maps" % (dmap.name, len(dmap.chain_maps))
        for chm in dmap.chain_maps :
            chm.display = False


    def StrucCloseChainMaps ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return
        
        print "%s - closing %d chain maps" % (dmap.name, len(dmap.chain_maps))

        for chm in dmap.chain_maps :
            chm.close()

        dmap.chain_maps = []



    def StrucDelChainMaps ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        print "%s - deleting %d chain maps" % (dmap.name, len(dmap.chain_maps))
        for chm in dmap.chain_maps :
            try : os.remove ( path + chm.name )
            except : print " - could not delete", chm.name
            chm.close()            

        dmap.chain_maps = []


    def StrucChMapsOvRegs ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        print "%s - %d chain maps" % (dmap.name, len(dmap.chain_maps))

        smod = self.CurrentSegmentation()
        if smod == None : return
        if len(smod.regions) == 0 : print " - no regions!"; return


        for mn in ["2AW4.pdb", "2AVY.pdb"] :
            m = getMod (mn)
            if m :
                for res in m.residues :
                    res.ribbonDisplay = False


        for chm in dmap.chain_maps :
            try : chm.display = False
            except : continue


        for chm in dmap.chain_maps :

            try : chm.display = False
            except : continue

            if chm.chain_id == "A" : continue
            if chm.chain_id == "B" and chm.mol.name == "2AW4.pdb" : continue

            chm.display = True
            print "%s - %s" % (chm.name, chm.chain_id)
            self.struc_map.set ( chm.name )

            oregs = self.ShowOverlappingRegions ()
            chimera.viewer.viewAll ()


            chm.display = False

            if len(oregs) == 0 :
                print " - NO overlapping regions"
                continue


            ress = chimera.selection.OSLSelection( "#%d:.%s" % (chm.mol.id, chm.chain_id) ).residues()
            for res in ress : res.ribbonDisplay = True

            mname = os.path.splitext ( chm.mol.name )[0]
            fname = mname + "_%s.pdb" % chm.chain_id
            fmol = getMod ( fname )
            if fmol == None :
                fmol = chimera.openModels.open ( path + fname )[0]
                fmol.ch_colors = RandColorChains ( fmol )

            self.struc.set ( fname )
            self.StrucCenter ()
            self.simRes.set ( "14" )
            self.simGridSp.set ( "2" )
            self.GenStrucMap ()
            sim_dmap = getMod ( self.struc_map.get() )
            sim_dmap.chmap = chm

            self.rotaSearch.set ( 0 )
            self.FitMapToSelRGroup ()
            rmsd = RMSD ( ress, fmol )
            print "\nRMSD: %f " % rmsd,
            tag = chm.chain_id

            msk_dmap = None
            
            if rmsd > 15.0 :
                msk_dmap = self.MaskMapWRegions ()
                self.FitMapToSelRGroup ()
                rmsd = RMSD ( ress, fmol )
                print "\nRMSD: %f " % rmsd,
                tag = chm.chain_id + "m"

            if rmsd > 15.0 :
                self.rotaSearch.set ( 1 )
                self.FitMapToSelRGroup ()
                rmsd = RMSD ( ress, fmol )
                print "\nRMSD: %f " % rmsd,
                tag = chm.chain_id + "r"

            if rmsd > 15.0 :
                self.rotaSearch.set ( 1 )
                self.FitMapToSelRGroup ( msk_dmap )
                rmsd = RMSD ( ress, fmol )
                print "\nRMSD: %f " % rmsd,
                tag = chm.chain_id + "rm"


            if rmsd < 15.0 :

                print "_________ YES _______\n"

                self.SelRegsOptimizeShapeScore ()
                oregs = self.ShowOverlappingRegions ()
                self.SelRegsOptimizeShapeScore ()
                oregs = self.ShowOverlappingRegions ()

                path = os.path.dirname ( dmap.data.path ) + os.path.sep
                log_file = path + "ribo_fits_sms_corr.txt"
                print "SMS log to", log_file
                fpl = open ( log_file, 'a' )
                fpl.write ( "%s %f " % (tag, sim_dmap.sms) )
                fpl.close ()

                if len(oregs) == 1 :
                    oregs[0].placed = True
                if len(oregs) >= 2 :
                    nreg = smod.join_regions ( oregs )

            else :

                print "_________ NO _______\n"


            #sim_dmap.close()
            sim_dmap.display = False

            if msk_dmap : msk_dmap.close()

            self.StrucCloseChainMap()

            #return
            

    def StrucCloseChainMap ( self ) :

        chm = getMod ( self.struc_map.get() )
        if chm :
            print " --- closing ", chm.mol.name
            chimera.openModels.close ( [chm.mol] )
            print " --- closing ", chm.chmap.name
            chm.chmap.close()
            print " --- closing ", chm.name
            chm.close()
            self.struc_map.set ( "" )




    def StrucBestShapeScore ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.fmap.get(), "not found"; return

        # first find out which region overlaps the most
        
        smod = self.CurrentSegmentation()
        if smod == None : return
        if len(smod.regions) == 0 : print " - no regions!"; return

        max_overlap_num = 0
        max_overlap_reg = None

        print "Region with highest overlap: ",

        for region in smod.regions :

            noverlap = 0
            for i,j,k in region.points() :
                try : noverlap += imap[k][j][i]
                except : continue

            if noverlap > max_overlap_num :
                max_overlap_num = noverlap
                max_overlap_reg = region

        if max_overlap_num == 0 :
            print "no region overlaps"
            return None

        else :
            print "%d (%d)" % (max_overlap_reg.rid, max_overlap_num)    
            sms = self.ShapeMatchScore ( fmap.fmol.atoms, dmap, [max_overlap_reg] )
            print " - shape match : %f" % sms

        return sms

    


    def SegAccuracy ( self, tag = "_acc", joinRegs=True ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return
        map_name = os.path.splitext ( dmap.name )[0]

        smod = self.CurrentSegmentation()
        if smod == None : return

        cmaps = dmap.chain_maps
        print "\nMapping %d regions -> %d chain maps..." % (
            len(smod.regions), len(cmaps) )

        if len(smod.regions) == 0 : print " - no regions!"; return
        if len(cmaps) == 0 : print " - no chain maps!"; return

        ch_regs = {}

        for reg_i, region in enumerate ( smod.regions ) :

            max_overlap_num = 0
            max_overlap_cmap = None

            ipoints = region.points()

            for cm in cmaps :

                noverlap = 0
                for i,j,k in ipoints :
                    try : noverlap += cm.imap[k][j][i]
                    except : continue

                if noverlap > max_overlap_num :
                    max_overlap_num = noverlap
                    max_overlap_cmap = cm

            if max_overlap_num == 0 :
                print "%d of %d - %d %d points - " % (reg_i+1, len(smod.regions), region.rid, len(ipoints) )

            else :
                # print "%d/%d - %d %d points - max overlap %d with %s" % (reg_i+1, len(smod.regions), region.rid, len(ipoints), max_overlap_num, max_overlap_cmap.name )
                c = max_overlap_cmap.surf_color
                region.color = c

                try : ch_regs[max_overlap_cmap].append ( region )
                except : ch_regs[max_overlap_cmap] = [ region ]

        thr = dmap.surface_levels[0]
        scores = []

        for cm in cmaps :
            print cm.name,
            if ch_regs.has_key ( cm ) == False :
                print " - no regions!"
                continue

            regs = ch_regs[cm]
            print " - %d regions" % len(regs),

            sms = self.ShapeMatchScore ( cm.atoms, dmap, regs )
            print " - sms %.4f" % (sms)

            if sms > 0.3 and len(regs) < 10 :
                scores.append ( sms )

            else :
                print "sms low: %.3f, nregs: %d" % (sms, len(regs))

            if joinRegs : smod.join_regions ( regs )


        map_name = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        log_f = path + map_name + tag + ".txt"
        print "\nAccuracies log file:", log_f

        res = NumberFromName ( dmap.name, 'r' );
        if res == None : res = 0.0
        print " - resolution: %.0f, %d scores" % ( res, len(scores) )

        try : fp = open ( log_f, "a" )
        except : print " - could not open log file"; return
        fp.write ( "%f %f %f %f " % ( res, min(scores), max(scores), sum(scores)/float(len(scores))) )
        for sm_score in scores : fp.write ( "%f " % sm_score )
        fp.write ( "\n" )
        fp.close ()




    def ShowChRegs ( self ) :

        mol = getMod ( self.struc.get() )
        if mol == None : print self.struc.get(), "not open"; return
        print "Structure:", mol.name

        try : chain_colors = mol.chain_colors
        except :  chain_colors = RandColorChains ( mol )
        mol.chain_colors = chain_colors

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        try : mol.chain_maps
        except : self.MakeChainMaps ( mol, dmap )
        print "- %d chain maps" % len ( mol.chain_maps )

        smod = self.CurrentSegmentation()
        if smod == None : return
        print "- %d regions" % len ( smod.surfacePieces )


        # for sp in smod.surfacePieces : sp.display = False

        rgroups = {}
        rgroups['all'] = []
        ov_regs = []

        for chmap in mol.chain_maps :

            # if chmap.chain_id == 'A' or chmap.chain_id == 'B' : continue
            # chmap.display = True

            regs = self.OverlappingRegions ( dmap, chmap, smod )

            ov_regs = ov_regs + regs

            if len(regs) == 0 :
                print " - no regions found";
                return

            # rgroups[chmap.chain_id] = regs

            #sms = self.ShapeMatchScore ( chmap, dmap, regs )
            #print " - chain %s, map overlaps %d regions - %.4f" % ( chmap.chain_id, len(regs), sms )

            for r in regs :
                #r.display = True
                #r.color = chmap.surf_color
                rgroups['all'].append ( r )

            # break

        if 0 :
            # for overlapping regs, which chain do they overlap the most
            for r in ov_regs :
                max_ov = 0.0
                max_ov_chm = None
                for chm in mol.chain_maps :
                    imap = self.MapIndexesInMap ( dmap, chm ) 
                    ipoints = r.points()
                    noverlap = 0
                    for i,j,k in ipoints :
                        if (i,j,k) in imap:
                            noverlap += 1

                    ov = float(noverlap) / r.point_count()
                    if ov > max_ov :
                        max_ov = ov
                        max_ov_chm = chm

                try : rgroups[max_ov_chm.chain_id]
                except : rgroups[max_ov_chm.chain_id] = []
                rgroups[max_ov_chm.chain_id].append ( r )

        return rgroups



    def SimChainMaps ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        map_name = os.path.splitext ( dmap.name ) [0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        res, grid = None, None

        try : res = float ( self.simRes.get() )
        except : print "Invalid number entered for resolution:", self.simRes.get(); return

        try : grid = float ( self.simGridSp.get() )
        except : grid = res / 3.0

        mol = getMod ( self.struc.get() )
        if mol == None : print "Structure", self.struc.get(), "not found"; return

        try : mol.chain_colors
        except : mol.chain_colors = RandColorChains ( mol )
            
        print "Simulating %d chain maps for %s, res %.3f, grid %.3f" % (
            len(mol.chain_colors.keys()), mol.name, res, grid)


        dmap.fitted_mols = []

        for cid, clr in mol.chain_colors.iteritems() :

            cname = map_name + "_" + cid + ".mrc"

            mv = getMod ( cname )

            if mv == None :

                sel_str = "#%d:.%s" % (mol.id, cid)
                print "%s [%s]" % (cname, sel_str),

                cmd = "molmap %s %f sigmaFactor 0.187 gridSpacing %f" % ( sel_str, res, grid )
                print " -", cmd
                chimera.runCommand ( cmd )

                for mod in chimera.openModels.list() :
                    ts = mod.name.split()
                    if len(ts) > 1 and mod.name.find("map") >=0 and mod.name.find("res") >=0 :
                        print " - saving to:", path + cname
                        mod.write_file ( path + cname, "mrc" )
                        chimera.openModels.close ( mod )
                        break

                mv = VolumeViewer.open_volume_file ( path + cname )[0]
                print " - loaded:", mv.name

            class FakeMolecule:
                def __init__(self, fmap):
                    self.fmap = fmap

            fmol = FakeMolecule ( mv )
            
            dmap.fitted_mols.append ( fmol )

            #gv.imap = imap
            #gv.name = cname
            #gv.chain_id = cid
            #dmap.chain_maps.append ( gv )

        


    def GenStrucMap ( self , show = True ) :

        res, grid, ress = None, None, None
        res_grid = (self.simRes.get(), self.simGridSp.get())

        if len(res_grid) >= 1 :
            rs = res_grid[0].split (":")
            if len(rs) == 3 :
                try :
                    first_res = float(rs[0])
                    incr_res = float(rs[1])
                    last_res = float(rs[2])
                except :
                    print "Invalid number entered for resolutions:", res_grid[0]
                    return

                ress = range ( first_res, last_res+1.0, incr_res )
                print "Resolutions:", ress
            else :
                try :
                    res = float ( rs[0] )
                except :
                    print "Invalid number entered for resolution:", res_grid[0]
                    return

        if len(res_grid) >= 2 and res_grid[1]:
            try :
                grid = float ( res_grid[1] )
            except :
                print "Invalid number for grid spacing:", res_grid[1]
                return

        
        if res == None and ress == None :
            print "Please enter a resolution"
            return

        elif res != None :
            if grid == None : grid = res/3.0
            dmap = self.DoGenStrucMap ( res, grid, show )
            if dmap == None: return
            self.struc_map.set ( dmap.name )


        elif ress != None :

            dmap = segmentation_map()
            if dmap == None : print "No segmentation map"; return

            start_thr = dmap.surface_levels[0]

            for r in ress :

                if grid == None : grid = r/3.0
                print "____________", r, "____________", grid, "____________"

                mv = self.DoGenStrucMap ( r, grid, show )

                print "Starting threshold: %f" % start_thr
                mv.surface_levels[0] = start_thr
                mv.update_surface ( False, VolumeViewer.volume.Rendering_Options() )

                try : dmap.chain_maps
                except : print "No chain maps for %s" % dmap.name; return

                if dmap.name == mv.name :
                    print "First model"
                else :
                    mv.chain_maps = []
                    print "Transferring %d chain maps from %s..." % ( len(dmap.chain_maps), dmap.name )
                    for cmap in dmap.chain_maps :
                        mv.chain_maps.append ( TransferChainMap (cmap, mv) )

                import segment_dialog
                d = segment_dialog.volume_segmentation_dialog()
                d.dmap.set ( mv.name )
                d.Segment ();
                self.SegAccuracy ( "_acc_max" )
                d.SmoothAndGroup ();
                self.SegAccuracy ( "_acc" )

                if dmap.name == mv.name :
                    print "First model"
                else :
                    for cm in mv.chain_maps : cm.close()
                    mv.close()

                smod = current_segmentation()
                if smod:
                    chimera.openModels.close ( [smod] )



    def DoGenStrucMap ( self, res, grid, show = True ) :

        mols = []
        if self.UseAllMods.get() :
            for mod in chimera.openModels.list() :
                if type(mod) == chimera.Molecule : mols.append ( mod )
        else :
            if len(self.struc.get()) == 0 :
                umsg ( "Please select a structure first" )
                return
            m = getMod ( self.struc.get() )
            if m == None :
                umsg ( "%s not open" % self.struc.get() )
                return
            mols.append ( m )

        if len(mols) == 0 :
            umsg ( "Structure not found" )
            return

        for m in mols:
            if not hasattr ( m, 'centered' ):
                centerMol ( m )

        #try : m.centered
        #except : self.StrucCenter ()

        base = ""
        sel_str = ""
        for m in mols :
            base = base + os.path.splitext(m.name)[0] + "_"
            sel_str = sel_str + "#%d," % m.id

        sel_str = sel_str [:-1]
        mname = base + "r%.1f_sp%.1f.mrc" % (res, grid)

        mv = getMod ( mname )
        if mv != None :
            print "Found", mname
            return mv
        
        print "Generating", mname

        cmd = "molmap %s %f sigmaFactor 0.187 gridSpacing %f" % ( sel_str, res, grid )
        print " -", cmd
        chimera.runCommand ( cmd )

        mv = None

        for mod in chimera.openModels.list() :
            ts = mod.name.split()
            if len(ts) > 1 and mod.name.find("map") >=0 and mod.name.find("res") >=0 :
                print " - found", mod.name
                #print " - saving to:", path + mname
                #mod.write_file ( path + mname, "mrc" )
                #print " - closing:", mod.name
                #chimera.openModels.close ( mod )
                mv = mod
                mv.name = mname
                break

        #mv = VolumeViewer.open_volume_file ( path + mname )[0]
        #print " - loaded:", mv.name
        mv.display = show
        clr = mols[0].color.rgba()
        mv.surfacePieces[0].color = ( clr[0], clr[1], clr[2], 1.0 )
        mv.mol = mols[0]

        return mv



    def StrucMapVolume ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        vol = _surface.enclosed_volume ( *(fmap.surfacePieces[0].geometry) )[0]
        print fmap.name + " volume: %f" % vol

        print " - surface levels",fmap.surface_levels
        thr = fmap.surface_levels[0]

        mm = fmap.data.matrix()
        mmab = numpy.where ( mm > thr, numpy.ones_like(mm), numpy.zeros_like(mm) )
        nz = numpy.shape ( numpy.nonzero ( mmab ) )[1]
        vvol = fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        tvol = vvol * float(nz)
        print "%s - %d above %f, volume %.3f" % (fmap.name, nz, thr, tvol)

        vvol = fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]

        tvol = vvol * float(nz)
        print " - %d above %f, volume %.3f" % (nz, thr, tvol)




    def FitSMapToDMap ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        print "Fitting %s to %s" % ( fmap.name, dmap.name )

        ov, corr = self.FitMapLocal ( fmap, dmap )

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        try : fmap.mol.chain_maps
        except : fmap.mol.chain_maps = []
        for chm in fmap.mol.chain_maps : chm.openState.xform = xfA


    def StrucBioMT ( self ) :

        if len(self.struc.get()) == 0 :
            umsg ( "Please select a structure first" )
            return
        mol = getMod ( self.struc.get() )
        if mol == None :
            umsg ( "%s not open" % self.struc.get() )
            return


        mats = BioMatrices ( mol )

        umsg ( "%s - %d matrices" % ( mol.name, len(mats) ) )




    def NextBioMt ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return False

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        try : mol = fmap.mol
        except : "no mol"; return False

        try : mol.bio_mts
        except : mol.bio_mts = BioMatrices ( mol )

        print " - %d bio matrices" % len(mol.bio_mts)

        try : mol.bio_mt_at = mol.bio_mt_at + 1
        except : mol.bio_mt_at = 1

        print "_______________ at matrix %d _______________" % mol.bio_mt_at

        try : amat = mol.bio_mts[ mol.bio_mt_at ]
        except : print " - no such matrix"; return False

        fmap.M = fmap.M0 * am_2_M(amat)

        # orthogonalize M
        T = fmap.M; xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
        mt, mr = xf_2_M ( xf )
        fmap.M = mt * mr

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        try : mol.chain_maps
        except : mol.chain_maps = []
        for chm in mol.chain_maps : chm.openState.xform = xfA

        return True



    def GoBioMt ( self ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return False
        try : mol = fmap.mol
        except : "no mol"; return False

        smod = self.CurrentSegmentation()
        if smod == None : return

        smod.bioM_groups = {}

        mol.bio_mt_at = 0

        for r in smod.surfacePieces :
            r.max_ov = 0.0
            r.max_ov_cid = None
            r.max_ov_bioM = None

        smod.p_ctrs = []            

        while self.NextBioMt () :

            self.FitSMapToDMap ()

            smod.bio_mt_at = mol.bio_mt_at
            smod.bioM_groups[mol.bio_mt_at] = self.ShowChRegs ()

            #self.ZeroDMap_with_FMap ()


            if 1 :
                atoms = chimera.selection.OSLSelection( "#%d:.%s" % (fmap.mol.id, 'P') ).atoms()
                points = _multiscale.get_atom_coordinates ( atoms, transformed = False )
                print " - %d points" % len(points)
                com = numpy.sum(points, axis=0) / len(points)
                comv = numpy.array ( [ [ com[0], com[1], com[2] ] ], numpy.float32 )
                _contour.affine_transform_vertices( comv, Matrix.xform_matrix(fmap.openState.xform) )
                _contour.affine_transform_vertices( comv, Matrix.xform_matrix(smod.openState.xform.inverse()) )

                print com[0], com[1], com[2]
                print comv[0][0], comv[0][1], comv[0][2]

                C = chimera.Vector ( comv[0][0], comv[0][1], comv[0][2] )
                smod.p_ctrs.append ( C )

            #return



    def ExtractProteins ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        if len(self.struc.get()) == 0 : umsg ("Please select a structure first"); return
        fmol = getMod ( self.struc.get() )
        if fmol == None : umsg ( "%s not open" % self.struc.get() ); return

        try : fmol.ch_colors
        except : fmol.ch_colors = RandColorChains ( fmol )

        mname = os.path.splitext ( fmol.name )[0]

        print "%s - %d chains" % (fmol.name, len(fmol.ch_colors.keys()))

        for cid, clr in fmol.ch_colors.iteritems () :

            print cid,
            cmol = copyMolChain ( None, fmol, cid, cid, None, clr.rgba() )
            cmol.name = "%s_%s.pdb" % (mname, cid)

            print " - %d atoms" % len(cmol.atoms)

            if len(cmol.atoms) > 0 :
                print " - writing", path + cmol.name
                chimera.PDBio().writePDBfile ( [cmol], path + cmol.name )


        print ""
        


    def SaveStrucFit ( self ) :
        
        if len ( self.struc_map.get() ) == 0 : umsg ("Please select a registered map first" ); return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return

        self.SaveFit ( fmap )


    def SaveFit ( self, fmap, clr=(rand(), rand(), rand(), 1) ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        try : fit_m_at = len ( dmap.fitted_mols ) + 1
        except : fit_m_at = 1; dmap.fitted_mols = []

        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        mol = CopyMol ( fmap.mol )
        mol.name = os.path.splitext ( mol.name )[0] + "_f%d.pdb" % fit_m_at

#        r = fmap.mol.residues[0]
#        c = r.ribbonColor
#        if c == None:
#            c = r.molecule.color
#        clr = c.rgba()
        clr = ( rand(), rand(), rand(), 1.0 )

        molApplyT ( mol, fmap.M )
        print "saving", path + mol.name

        chimera.PDBio().writePDBfile ( [mol], path + mol.name )
        m = chimera.openModels.open ( path + mol.name )[0]
        print "opened", m.name
        for res in m.residues :
            res.ribbonDisplay = True
            res.ribbonDrawMode = 2
            res.ribbonColor = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 );
            for at in res.atoms :
                at.display = False

        if dmap : dmap.fitted_mols.append ( m )

        m.display = True

        umsg ( "Saved %s" % ( path + mol.name ) )

        if 0 :
            m.fmap = fmap.copy ()
            m.fmap.name = os.path.splitext (m.name)[0] + ".mrc"
            print "Saved fit map %s (thr %f)" % ( m.fmap.name, m.fmap.surface_levels[0] )


        return m
        


    def FitToSelRegion ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        smod = self.CurrentSegmentation()
        if smod is None : return

        fmap = self.SimulatedMap()
        if fmap == None:
            return

        regs = smod.selected_regions()
        if len(regs) != 1 :
            umsg ( "Please select one region to register the structure with" )
            return

        self.FitToEachRegion ( smod, fmap, dmap, regs, 1 )



    def FitToSelectedRegions ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        smod = self.CurrentSegmentation()
        if smod is None : return

        fmap = self.SimulatedMap()
        if fmap == None:
            return

        regs = smod.selected_regions()
        if len(regs) == 0 :
            #umsg ( "Please select one or more regions to register the structure with" )
            #return
            regs = smod.regions
            print "Fitting to all regions"

        self.FitToEachRegion ( smod, fmap, dmap, regs, 0 )



    def FitToEachRegion ( self, smod, fmap, dmap, regs, show=1 ) :

        if not hasattr(fmap.mol, 'centered'):
            self.StrucCenter()

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        reportFitRegions(fmap.name, regs)
        scores = []
        corrs = []

        sregs = set ( smod.regions )

        for reg in regs :

            if not reg in sregs:
                continue
            sp = reg.surface_piece
            sp.display = True
            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

            fmap.display = False
            fmap.mol.display = True

            self.FitMapToRegion ( reg )


            if 0 :
                #fmap.sms_score = self.OptShapeScore ( fmap, dmap, [ reg ] )
                fmap.sms_score = self.ShapeMatchScore ( fmap.mol.atoms, dmap, [reg] )
                # fmap.fit_score = correlation ( fmap, dmap, True )

            scores.append ( 0 )
            corrs.append ( fmap.fit_score )

            if self.strucCopyAfterFit.get() :
                reg.placed = self.SaveFit ( fmap, reg.color )

            umsg ( "Cross-correlation of fit: %f" % fmap.fit_score )


        res = NumberFromName ( fmap.name, 'r' )
        if res == None : res = 0.0

        acc_log_file = path + smod.name + "_fits_acc.txt"
        print "Accuracies log to", acc_log_file

        try : fp = open ( acc_log_file, "a" )
        except : print " - could not open log file"; return

        fp.write ( "%f %f %f %f " % ( res, min(scores), max(scores), sum(scores)/float(len(scores))) )
        for sm_score in scores : fp.write ( "%f " % sm_score )
        fp.write ( "\n" )

        fp.write ( "C%f %f %f %f " % ( res, min(corrs), max(corrs), sum(corrs)/float(len(corrs))) )
        for sm_score in corrs : fp.write ( "%f " % sm_score )
        fp.write ( "\n" )

        fp.close ()



    def FitMapToRegion ( self, sreg ) :

        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        tpoints = sreg.map_points()
        if self.rotaSearch.get () :
            self.saFitMapToPoints_byRot ( fmap, tpoints, dmap )
        else :
            self.saFitMapToPoints ( fmap, tpoints, dmap )





    def saFitMapToPoints ( self, fmap, points, dmap ) :

        print "fitting %s in map %s, to %d points" % (fmap.name, dmap.name, len(points))
        COM, U, S, V = prAxes ( points )

        # mT = self.CSegMol
        highest_fit, highest_M, highest_MI, opt_af = 0.0, None, None, None

        comT = numpy.matrix ( [
            [ 1, 0, 0, COM[0] ],
            [ 0, 1, 0, COM[1] ],
            [ 0, 0, 1, COM[2] ],
            [ 0, 0, 0,      1 ]  ] )

        tXO, tXR = xf_2_M ( dmap.openState.xform )

        fmap.fpoints = VolumeData.grid_indices ( fmap.data.size, numpy.single)        # i,j,k indices
        _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
        fmap.fpoint_weights = numpy.ravel( fmap.data.matrix() ).astype(numpy.single)

        threshold = fmap.surface_levels[0]
        #threshold = .3 * max ( fmap.fpoint_weights )

        ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
        fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
        fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
        nz = numpy.nonzero( fmap.fpoint_weights )[0]

        print " - threshold %.4f, %d nonzero, center %s" % ( threshold, len(nz), str(COM) )

        if len(nz) < len (fmap.fpoint_weights) :
            fmap.fpoints = numpy.take( fmap.fpoints, nz, axis=0 )
            fmap.fpoint_weights = numpy.take(fmap.fpoint_weights, nz, axis=0)

        mass = numpy.sum(fmap.fpoint_weights, dtype=numpy.single)
        fmap.rotation_center = numpy.dot(fmap.fpoint_weights,fmap.fpoints) / mass


        # the 4 alignments to try...
        amult = [ [1.0,1.0,1.0], [-1.0,-1.0,1.0], [-1.0,1.0,-1.0], [1.0,-1.0,-1.0] ]

        for j in range(4) :

            af = amult[j]
            print " ~ %.1f*X %.1f*Y %.1f*Z" % ( af[0], af[1], af[2] )

            mR = numpy.matrix ( [
                [ af[0]*U[0,0], af[1]*U[0,1], af[2]*U[0,2], 0 ],
                [ af[0]*U[1,0], af[1]*U[1,1], af[2]*U[1,2], 0 ],
                [ af[0]*U[2,0], af[1]*U[2,1], af[2]*U[2,2], 0 ],
                [            0,            0,            0, 1 ] ] )

            fmap.M = comT * mR
            T = tXO * tXR * fmap.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA

            # if j == 0 : return

            ov, corr = self.FitMap_T ( fmap, dmap, num_tries=4 )

            # ov, corr = map_overlap_and_correlation ( fmap, dmap, True )
            #print "- final overlap: %f, correlation: %f\n" % (ov, corr)
            #mA.avD, n = self.AvgDVal ( dmap, mA )

            if corr > highest_fit :
                highest_fit = corr
                highest_M = fmap.M
                highest_MI = comT * mR
                opt_af = af

        print " * highest fit: %f for %.1f*X %.1f*Y %.1f*Z" % (
            highest_fit, opt_af[0], opt_af[1], opt_af[2] )

        fmap.M = highest_M
        fmap.MI = highest_MI
        fmap.fit_score = highest_fit

        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA


        if 0 :
            f_m = fmap.data.full_matrix(); size = list(f_m.shape); size.reverse()
            fmap.fpoints = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
            _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
            fmap.fpoint_weights = numpy.ravel(f_m).astype(numpy.single)

            threshold = fmap.surface_levels[0]
            ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
            fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
            fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
            nz = numpy.nonzero( fmap.fpoint_weights )[0]
            print " - threshold %.4f, %d nonzero" % ( threshold, len(nz) ), COM

            ov, corr = self.FitMap_T ( fmap, dmap, num_tries=2 )
            #ov, corr = map_overlap_and_correlation ( fmap, dmap, True )
            print " - correlation with full threshold: %f" % corr
            fmap.fit_score = corr

            T = tXO * tXR * fmap.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA
            fmap.mol.openState.xform = xfA



    def saFitMapToPoints_byRot ( self, fmap, points, dmap, N=8, M=8 ) :

        fpLog = None
        if 0 : fpLog = open ( "fit_rot_log.txt", "w" )

        print "fitting %s in map %s, to %d points, by rotation" % (fmap.name, dmap.name, len(points))
        COM, U, S, V = prAxes ( points )

        comT = numpy.matrix ( [
            [ 1, 0, 0, COM[0] ],
            [ 0, 1, 0, COM[1] ],
            [ 0, 0, 1, COM[2] ],
            [ 0, 0, 0,      1 ]  ] )

        tXO, tXR = xf_2_M ( dmap.openState.xform )

        f_m = fmap.data.full_matrix(); size = list(f_m.shape); size.reverse()
        fmap.fpoints = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
        _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
        fmap.fpoint_weights = numpy.ravel(f_m).astype(numpy.single)

        #threshold = fmap.surface_levels[0]
        threshold = .3 * max ( numpy.ravel(f_m).astype(numpy.single) )

        ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
        fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
        fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
        nz = numpy.nonzero( fmap.fpoint_weights )[0]

        print " - threshold %.4f, %d nonzero" % ( threshold, len(nz) ), COM

        if len(nz) < len (fmap.fpoint_weights) :
            fmap.fpoints = numpy.take( fmap.fpoints, nz, axis=0 )
            fmap.fpoint_weights = numpy.take(fmap.fpoint_weights, nz, axis=0)

        mass = numpy.sum(fmap.fpoint_weights, dtype=numpy.single)
        fmap.rotation_center = numpy.dot(fmap.fpoint_weights,fmap.fpoints) / mass


        # dmap = self.DoMask ( dmap, mol_set )

        highest_fit, highest_M, highest_MI, opt_af = 0.0, None, None, None

        # tt = self.fitByRot_NM.get().split(',')
        # N, M = int ( tt[0] ), int ( tt[1] )
        # N, M = 10, 10
        print "N %d, M %d" % (N, M)


        thetas, phis = [], []
        # http://www.math.niu.edu/~rusin/known-math/97/spherefaq


        for k in range ( 1, N+1 ) :
            h = -1.0 + ( 2.0*float(k-1)/float(N-1) )
            thetas.append ( numpy.arccos(h) )
            phi = 0.0
            if k == 1 or k == N : phi = 0
            else : phi = (phis[k-2] + 3.6/numpy.sqrt(float(N)*(1.0-h**2.0)))
            while phi > 2.0*numpy.pi : phi -= 2.0*numpy.pi
            phis.append ( phi )


        for ti, thet in enumerate ( thetas ) :
            phi, theta = phis[ti], thet - (numpy.pi / 2.0)
            v = chimera.Vector (0,0,0)
            v[2] = numpy.sin(theta)
            v[1] = numpy.cos(theta) * numpy.sin (phi)
            v[0] = numpy.cos(theta) * numpy.cos (phi)

            print "%d/%d : %.3f (%.3f), %.3f %.3f, %f" % (
                ti, len(thetas), theta, theta*180.0/numpy.pi, phi, phi*180.0/numpy.pi, v.length )

            for m in range ( M ) :

                ang = 360.0 * float(m)/float(M)
                print "- %d/%d - %.3f deg" % (m, M, ang)

                status ( "Rotational fit %d/%d" % (ti*M+m+1, N*M) )

                xfR = chimera.Xform.rotation ( v, ang )
                rrO, rrR = xf_2_M ( xfR )

                fmap.M = comT * rrR
                T = tXO * tXR * fmap.M
                xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
                fmap.openState.xform = xfA
                fmap.mol.openState.xform = xfA

                ov, corr = self.FitMap_T ( fmap, dmap, num_tries=1 )

                # ov, corr = map_overlap_and_correlation( fmap, dmap, True )
                #print "- corr %f" % corr

                if fpLog : fpLog.write ( "%f\n" % corr )

                if corr > highest_fit :
                    highest_fit = corr
                    highest_M = fmap.M
                    highest_MI = comT * rrR


        print "- highest fit: %f" % ( highest_fit )
        print ""

        if fpLog : fpLog.close()
        

        fmap.M = highest_M
        fmap.MI = highest_MI
        fmap.fit_score = highest_fit

        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA





    def FitMap_T ( self, fmap, dmap, Log=False, num_tries=1, bTrans=True, bRot=True ) :

        fpLog = None
        if 0 :
            try : self.fitno = self.fitno + 1
            except : self.fitno = 0
            fpLog = open ( "fit_log_%d.txt" % self.fitno, "a" )

        olap, cor = 0, 0

        for gd_i in range ( num_tries ) :

            #print "%d: fitting of %s in %s" % (gd_i, fmap.name, dmap.name)
            print "   try %d of %d:" % (gd_i+1, num_tries),

            T = fmap.M; data_xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            data_xyz_to_xyz = Matrix.xform_matrix( data_xf )

            xyz_to_ijk = dmap.data.xyz_to_ijk_transform
            ttf = Matrix.multiply_matrices(xyz_to_ijk, data_xyz_to_xyz)

            data_array = dmap.data.matrix()
            max_steps = 100
            ijk_step_size_max = 0.5
            ijk_step_size_min = 0.01
            optimize_translation, optimize_rotation = True, True

            ##### from FitMap.locate_maximum ####

            segment_steps = 4
            cut_step_size_threshold = .5
            step_cut_factor = .5

            ijk_step_size = ijk_step_size_max

            move_tf = Matrix.identity_matrix()

            step = 0
            while step < max_steps and ijk_step_size > ijk_step_size_min:
                xyz_to_ijk_tf = Matrix.multiply_matrices( ttf, move_tf )

                try :
                    seg_tf = FitMap.step_to_maximum(
                        fmap.fpoints, fmap.fpoint_weights,
                                             data_array, xyz_to_ijk_tf,
                                             segment_steps, ijk_step_size,
                                             bTrans, bRot,
                                             fmap.rotation_center)
                except :
                    seg_tf = FitMap.step_to_maximum(
                        fmap.fpoints, fmap.fpoint_weights,
                                             data_array, xyz_to_ijk_tf,
                                             segment_steps, ijk_step_size,
                                             bTrans, bRot,
                                             fmap.rotation_center,
                                             'sum product' )
                    
                step += segment_steps
                mm = FitMap.maximum_ijk_motion(
                    fmap.fpoints, xyz_to_ijk_tf, seg_tf)
                if mm < cut_step_size_threshold * segment_steps * ijk_step_size:
                    ijk_step_size *= step_cut_factor
                move_tf = Matrix.multiply_matrices(seg_tf, move_tf)

                #shift, angle = FitMap.shift_and_angle(seg_tf, fmap.rotation_center)
                #print '- %d steps: shift %.3g, rotation %.3g degrees, step %.3f' % (
                #    step, shift, angle, ijk_step_size)
                # print "^",
                # add move_tf to overall transform


                if fpLog :
                    map_values, outside = VolumeData.interpolate_volume_data(
                        fmap.fpoints, xyz_to_ijk_tf, data_array )
                    olap, cor = overlap_and_correlation ( fmap.fpoint_weights, map_values )

                    fpLog.write ( "%f %f\n" % (olap, cor) )


                #if Log :
                #    avgd, n = self.AvgDVal ( dmap, mol )
                #    self.AvgDValue.set ( "%f" % avgd )

                #    log_f = self.path.get() + "_%s_%s_ConvergeMOIAlign.txt" % (dmap.name, mol.name)
                #    print "Log file: %s" % log_f
                #    f = open ( log_f, 'a' )
                #    f.write ( "%7d %7d %15f\n" % (9, step, avgd) )
                #    f.close()


            xT, xR = xf_2_M ( Matrix.chimera_xform ( move_tf ) )
            fmap.M = fmap.M * xT * xR

            map_values, outside = VolumeData.interpolate_volume_data(
                fmap.fpoints, xyz_to_ijk_tf, data_array )

            olap, cor = overlap_and_correlation ( fmap.fpoint_weights, map_values )

            try : shift, angle = FitMap.shift_and_angle( move_tf, fmap.rotation_center )
            except : shift, angle = Matrix.shift_and_angle( move_tf, fmap.rotation_center )
                
            # print ""
            print ' - %d steps: d %.3g, r %.3g, ov %.3e, cor %f' % (
                step, shift, angle, olap, cor )

            if angle < .1 and shift < .1 :
                break

        # tXO, tXR = xf_2_M ( dmap.openState.xform )
        # T = tXO * tXR * fmap.M
        # xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        # fmap.openState.xform = xfA

        if fpLog : fpLog.close()

        return olap, cor


    def FitMapLocal ( self, fmap, dmap ) :

        f_m = fmap.data.full_matrix(); size = list(f_m.shape); size.reverse()
        fmap.fpoints = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices
        _contour.affine_transform_vertices( fmap.fpoints, fmap.data.ijk_to_xyz_transform )
        fmap.fpoint_weights = numpy.ravel(f_m).astype(numpy.single)

        threshold = fmap.surface_levels[0]
        #threshold = .3 * max ( numpy.ravel(f_m).astype(numpy.single) )

        ge = numpy.greater_equal(fmap.fpoint_weights, threshold)
        fmap.fpoints = numpy.compress(ge, fmap.fpoints, 0)
        fmap.fpoint_weights = numpy.compress(ge, fmap.fpoint_weights)
        nz = numpy.nonzero( fmap.fpoint_weights )[0]

        print " - threshold %.4f, %d nonzero" % ( threshold, len(nz) )

        if len(nz) < len (fmap.fpoint_weights) :
            fmap.fpoints = numpy.take( fmap.fpoints, nz, axis=0 )
            fmap.fpoint_weights = numpy.take(fmap.fpoint_weights, nz, axis=0)

        mass = numpy.sum(fmap.fpoint_weights, dtype=numpy.single)
        fmap.rotation_center = numpy.dot(fmap.fpoint_weights,fmap.fpoints) / mass

        olap, cor = self.FitMap_T ( fmap, dmap, num_tries=1 )

        fmap.fpoints = None
        fmap.fpoint_weights = None
        fmap.rotation_center = None

        return olap, cor






    def StrucGroupRegions ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        smod = self.CurrentSegmentation()
        if smod == None : return


        regs = smod.selected_regions()

        if len(regs)==0 :
            print "\nGrouping all regions"
            tvol = self.MapVolume ( fmap )
            smod.rgroups = self.GroupAllRegions ( smod, tvol )

        elif len(regs) == 1 :
            tvol = self.MapVolume ( fmap )
            bRad = -1.0 # self.MapBoundingRad ( fmap )

            print "\nMaking groups around region %d - target vol %.3f, b-Rad %.3f" % (regs[0].rid, tvol, bRad)
            smod.rgroups, maxDepthReached = self.GroupAroundReg ( smod, regs[0], tvol, bRad )
            print " - depth reached: %d" % maxDepthReached


        else :
            print "Please select no regions (for global groups) or one region (for local groups)"
            return

        if len ( smod.rgroups ) == 0 :
            print "No groups result!"
            return

        smod.rgroups.sort()
        smod.rgroup_at = 0
        dv, regs = smod.rgroups[smod.rgroup_at]
        print "Group %d of %d - dv %f, regions:" % (smod.rgroup_at+1, len(smod.rgroups), dv),
        for r in regs : print r.rid,
        print ""

        for sp in smod.surfacePieces :
            if regs.count ( sp.region ) > 0 : sp.display = True
            else : sp.display = False




    def GroupAroundReg ( self, smod, reg, target_volume, bRad=-1.0 ) :

        dv_rgroups = []
        maxDepthReached = 0

        stack = [ [[reg], reg.enclosed_volume(), 0] ]

        print "Making groups around %d" % reg.rid,
        status ( "Making groups around %d" % reg.rid ), 

        while len(stack) > 0 :

            regs, vol_at, depth_at = stack.pop()

            if depth_at > maxDepthReached : maxDepthReached = depth_at

            if depth_at >= SAF_LS_DEPTH : continue
            if vol_at >= target_volume * (1.0 + SAF_DVOL)  : continue

            dv = abs ( vol_at - target_volume ) / target_volume
            dv_rgroups.append ( [dv, regs] )

            if len(dv_rgroups) % 100 == 0 :
                status ( "Making groups around %d - %d groups" % (reg.rid, len(dv_rgroups)) ), 

            reg_at = regs[0]
            for cr in reg_at.contacting_regions():
                if regs.count ( cr ) != 0 : continue
                if cr.placed : continue
                vol = vol_at + cr.enclosed_volume()
                stack.insert ( 0, [ [cr]+regs, vol, depth_at+1 ] )

        dv_rgroups_f = self.FilterGroups ( dv_rgroups, bRad )

        print " - %d groups --> %d filtered groups" % (len(dv_rgroups), len(dv_rgroups_f))
        return [dv_rgroups_f, maxDepthReached]

        



    def FilterGroups ( self, dv_rgroups, bRad = -1.0 ) :

        dv_rgroups_f = []
        len_groups = {}
        #          [ Included, ri ]
        inc_regs_map = {}

        gi, ngroups = 0, len(dv_rgroups)

        # print "Filtering %d groups..." % ngroups

        for dv, regs in dv_rgroups :

            gi = gi + 1
            if gi % 100 == 0 :
                status ( "Filtering group %d/%d" % (gi,ngroups) )

            if dv > SAF_DVOL : continue

            included = False

            regs.sort()
            inc_regs_at = inc_regs_map
            for reg in regs :
                try : included, inc_regs_at = inc_regs_at[reg]
                except : included = False; break

            #num_regs = len(regs)
            #try : i_dv_groups = len_groups [ num_regs ]
            #except : i_dv_groups = None; len_groups [ num_regs ] = []
            #if i_dv_groups != None :
            #    for idv, iregs in i_dv_groups :
            #        bAllIn = True
            #        for r in regs :
            #            if iregs.count ( r ) == 0 : bAllIn = False; break
            #        if bAllIn == True : included = True; break

            if included : continue

            if bRad > 0.0 :
                regs_bRad = regions.regions_radius(regs)

                brad_d = abs ( regs_bRad - bRad ) / bRad

                if brad_d > SAF_DBRAD : continue

            dv_rgroups_f.append ( [dv, regs] )

            #len_groups[ num_regs ].append ( [dv,regs] )

            inc_regs_at = inc_regs_map
            last_arr = None
            for reg in regs :
                try :
                    last_arr = inc_regs_at[reg]
                    inc_regs_at = last_arr[1]
                except :
                    last_arr = [ False, {} ]
                    inc_regs_at[reg] = last_arr
                    inc_regs_at = last_arr[1]

            last_arr[0] = True


        return dv_rgroups_f




    def GroupAllRegions ( self, smod, target_volume, bRad=-1.0) :

        dv_rgroups = []
        maxDepthReached = 0

        print "Grouping %d regions in %s, target volume %.2f, bounding radius %.2f" % (
            len(smod.surfacePieces), smod.name, target_volume, bRad )

        ri, nregs = 0, len(smod.regions)

        for reg in smod.regions :

            ri = ri + 1
            # status ( "Grouping around region %d/%d" % (ri,nregs) )

            dv_rgroupsR, maxDepthReachedR = self.GroupAroundReg ( smod, reg, target_volume, bRad )

            dv_rgroups = dv_rgroups + dv_rgroupsR

            if maxDepthReachedR > maxDepthReached : maxDepthReached = maxDepthReachedR

            
        print "\n - max depth reached: %d" % maxDepthReached

        print " - filtering %d groups..." % len( dv_rgroups )
        return self.FilterGroups ( dv_rgroups, bRad )
        #return self.FilterGroups ( dv_rgroups )





    def NextRGroup ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        smod.rgroup_at = smod.rgroup_at + 1

        if smod.rgroup_at >= len(smod.rgroups) : smod.rgroup_at = 0

        dv, regs = smod.rgroups[smod.rgroup_at]

        print "Group %d/%d - dv %f, regions:" % (smod.rgroup_at+1, len(smod.rgroups), dv),
        for r in regs : print r.rid,
        print ""

        for sp in smod.surfacePieces :
            if regs.count ( sp.region ) > 0 : sp.display = True
            else : sp.display = False



    def FindGroupFromSelRegs ( self ) :

        smod = self.CurrentSegmentation()
        if smod == None : return

        print "\nStructure:",
        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return
        tvol = self.MapVolume ( fmap )


        regs = smod.selected_regions()        
        if len(regs)==0 : print "no selected regions found"; return

        print "Finding group for %d selected regions:" % ( len(regs) ),
        regs_vol = 0.0
        for r in regs :
            print r.rid,
            regs_vol = regs_vol + r.enclosed_volume()
        print ""

        dv = abs ( tvol - regs_vol ) / tvol;
        print " - total regions volume: %.3f - DV %.5f" % (regs_vol, dv)

        points = numpy.concatenate ( [r.map_points()
                                      for r in regs], axis=0 )

        com = numpy.sum(points, axis=0) / len(points)
        C = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points = points - comv
        regs_bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )

        print " - COM: %.3f %.3f %.3f, bounding rad %.3f" % ( C[0], C[1], C[2], regs_bRad )


        print "Searching %d groups" % len(smod.rgroups)
        gi = 0
        for dv, gregs in smod.rgroups :
            gi = gi + 1
            if len(regs) != len (gregs) : continue
            allIn = True
            for r in regs :
                if gregs.count ( r ) == 0 : allIn = False; break
            if allIn :
                print " - FOUND! %d - dv %f - regs" % (gi, dv),
                for r in gregs : print r.rid,
                print ""

        print " - done searching"



    def FitMapToSelRGroup ( self, dmap = None ) :

        if dmap is None:
            dmap = segmentation_map()
            if dmap == None : print "No segmentation map"; return

        fmap = self.SimulatedMap()
        if fmap == None : return
        thr = fmap.surface_levels[0]
        mm = fmap.data.matrix()
        mmab = numpy.where ( mm > thr, numpy.ones_like(mm), numpy.zeros_like(mm) )
        nz = numpy.shape ( numpy.nonzero ( mmab ) )[1]
        tvol = float(nz) * fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        print "%s - %d above %f, volume %.3f" % (fmap.name, nz, thr, tvol)


        smod = self.CurrentSegmentation()
        if smod == None : return

        regs = smod.selected_regions()
        if len(regs)==0 : print "no selected regions found"; return

        reportFitRegions(fmap.name, regs)

        for reg in regs:
            clr = reg.color
            reg.surface_piece.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
        print ""

        points = numpy.concatenate ( [r.map_points()
                                      for r in regs], axis=0 )
        regions_vol = sum([r.enclosed_volume() for r in regs])

        dv = abs(regions_vol - tvol) / tvol
        print " - regions volume: %.2f - dv %.5f" % (regions_vol, dv)

        # self.saFitMapToPoints ( fmap, points, dmap )

        if self.rotaSearch.get () :
            self.saFitMapToPoints_byRot ( fmap, points, dmap )
        else :
            self.saFitMapToPoints ( fmap, points, dmap )


        #fmap.sms = self.ShapeMatchScore ( fmap.mol.atoms, dmap, regs )
        fmap.sms = 0 # self.OptShapeScore ( fmap, dmap, regs )
        #print " - shape match score: %.3f" % fmap.sms

        if self.strucCopyAfterFit.get() :
            self.SaveFit ( fmap, regs[0].color )

        umsg ( "Cross-correlation: %.4f\n" % (fmap.fit_score) )

        #dmap_name = os.path.splitext ( dmap.name )[0]
        #path = os.path.dirname ( dmap.data.path ) + os.path.sep
        #log_f = path + "%s_fits.txt" % (dmap_name)
        #print "Log file: %s" % log_f

        #fp = open ( log_f, 'a' )
        #fp.write ( "%s %f %f %f" % (fmap.mol.name, fmap.fit_score, fmap.sms, dv) )
        #T = fmap.M ; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
        #T = fmap.MI; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
        #for r in regs : fp.write ( " %d" % r.rid )
        #fp.write ( "\n" )
        #fp.close()



    def ShapeMaskedCorr ( self, fmap, dmap, regs ) :

        points = numpy.concatenate ( [r.map_points()
                                      for r in regs], axis=0 )

        #tf = Matrix.xform_matrix ( dmap.openState.xform )
        #_contour.affine_transform_vertices(points, tf)

        sg = VolumeData.zone_masked_grid_data ( dmap.data, points, 0.5 )

        #gv = VolumeViewer.volume.add_data_set ( sg, None )
        #gv.name = 'regions_mask'
        #gv.openState.xform = dmap.openState.xform

        nz = numpy.shape ( numpy.nonzero ( sg.matrix() ) )[1]
        if len(points) != nz :
            print "mask failed [%d points, %d masked]" % ( len(points), nz )
            return 0.0

        r_mask_matrix = sg.matrix()

        fmapm = fmap.data.matrix()
        f_weights = numpy.ravel(fmapm).astype(numpy.single)

        size = list(fmapm.shape)
        size.reverse()
        f_points = VolumeData.grid_indices(size, numpy.single)        # i,j,k indices

        thr = fmap.surface_levels[0]
        ge = numpy.greater_equal(f_weights, thr)
        f_points = numpy.compress(ge, f_points, 0)
        f_weights = numpy.compress(ge, f_weights)

        from Matrix import multiply_matrices as MM
        from Matrix import xform_matrix as XFM

        p2m_transform = MM ( dmap.data.xyz_to_ijk_transform,
                        MM ( XFM ( dmap.openState.xform.inverse() ),
                        MM ( XFM ( fmap.openState.xform ), fmap.data.ijk_to_xyz_transform ) ) )

        #_contour.affine_transform_vertices(f_points, tf)

        m_weights, outside = VolumeData.interpolate_volume_data (
            f_points, p2m_transform, r_mask_matrix, 'linear' )

        o, shape_cor = overlap_and_correlation ( f_weights, m_weights )

        print " * shape map-masked correlation %.3f" % shape_cor

        return shape_cor



    def OptShapeScore ( self, fmap, dmap, regs ) :

        thr_at = fmap.surface_levels[0]
        sms = self.ShapeMatchScore ( fmap, dmap, regs )

        umsg ( "Optimizing threshold for %s - %.2f/%.4f" % (fmap.name, thr_at, sms) )

        while 1 :
            fmap.surface_levels[0] = thr_at + .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )

            if new_sms > sms :
                thr_at = fmap.surface_levels[0]
                sms = new_sms
                #print "- %.2f/%.4f" % (thr_at, sms),
                
                ro = VolumeViewer.volume.Rendering_Options()
                fmap.update_surface ( False, ro )
                for sp in fmap.surfacePieces :
                    v, t = sp.geometry
                    if len(v) == 8 and len(t) == 12 : sp.display = False

                umsg ( "Optimizing threshold for %s - %.2f/%.4f" % (fmap.name, thr_at, sms) )

            else :
                fmap.surface_levels[0] = thr_at
                break

        while 1 :
            fmap.surface_levels[0] = thr_at - .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )
            if new_sms > sms :
                thr_at = fmap.surface_levels[0]
                sms = new_sms
                print "- %.2f/%.4f" % (thr_at, sms),

                ro = VolumeViewer.volume.Rendering_Options()
                fmap.update_surface ( False, ro )
                for sp in fmap.surfacePieces :
                    v, t = sp.geometry
                    if len(v) == 8 and len(t) == 12 : sp.display = False

                umsg ( "Optimizing threshold for %s - %.2f/%.4f" % (fmap.name, thr_at, sms) )

            else :
                fmap.surface_levels[0] = thr_at
                break

        #print "| %.2f/%.4f" % (thr_at, sms)
        return sms



    def OptShapeScoreRegs ( self, fmap, dmap, regs ) :

        thr_at = dmap.surface_levels[0]
        sms = self.ShapeMatchScore ( fmap, dmap, regs )

        print ", optz regs - %.2f/%.4f" % (thr_at, sms),

        while 1 :
            dmap.surface_levels[0] = thr_at + .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )
            if new_sms > sms :
                thr_at = dmap.surface_levels[0]
                sms = new_sms
                print "- %.2f/%.4f" % (thr_at, sms),
            else :
                break

            ro = VolumeViewer.volume.Rendering_Options()
            dmap.update_surface ( False, ro )
            for sp in dmap.surfacePieces :
                v, t = sp.geometry
                if len(v) == 8 and len(t) == 12 : sp.display = False

        while 1 :
            dmap.surface_levels[0] = thr_at - .01
            new_sms = self.ShapeMatchScore ( fmap, dmap, regs )
            if new_sms > sms :
                thr_at = dmap.surface_levels[0]
                sms = new_sms
                print "- %.2f/%.4f" % (thr_at, sms),
            else :
                break

            ro = VolumeViewer.volume.Rendering_Options()
            dmap.update_surface ( False, ro )
            for sp in dmap.surfacePieces :
                v, t = sp.geometry
                if len(v) == 8 and len(t) == 12 : sp.display = False

        dmap.surface_levels[0] = thr_at
        print "| %.2f/%.4f" % (thr_at, sms)
        return sms



    def ShapeMatchScore ( self, atoms, dmap, regs, bPrint=False ) :

        #fmol = fmap.mol
        #print "atoms from", fmol.name
        #points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = True )

        print "shape match of %d atoms" % len(atoms)
        points = _multiscale.get_atom_coordinates ( atoms, transformed = True )

        tfd = Matrix.xform_matrix( dmap.openState.xform.inverse() )
        #tff = Matrix.xform_matrix( fmap.openState.xform.inverse() )
        #tf = Matrix.multiply_matrices( tff, tfd )
        _contour.affine_transform_vertices( points, tfd )

        sg_fmap = VolumeData.zone_masked_grid_data ( dmap.data, points, max(3.0,dmap.data.step[0]) )

        #try : gv = VolumeViewer.volume.add_data_set ( sg_fmap, None )
        #except : gv = VolumeViewer.volume.volume_from_grid_data ( sg_fmap )
        #gv.name = 'registered_structure_region'
        #gv.openState.xform = dmap.openState.xform


        points = numpy.concatenate ( [r.map_points() for r in regs], axis=0 )

        sg_dmap = VolumeData.zone_masked_grid_data ( dmap.data, points, dmap.data.step[0] / 2.0 )
        #try : gv = VolumeViewer.volume.add_data_set ( sg_dmap, None )
        #except : gv = VolumeViewer.volume.volume_from_grid_data ( sg_dmap )
        #gv.name = 'regions_mask'

        regs_m = sg_dmap.matrix()
        regs_f = sg_fmap.matrix()
        regs_mz = numpy.where ( regs_m > dmap.surface_levels[0], numpy.ones_like(regs_m), numpy.zeros_like(regs_m) )
        regs_fz = numpy.where ( regs_f > dmap.surface_levels[0], numpy.ones_like(regs_f), numpy.zeros_like(regs_f) )
        
        nz = numpy.shape ( numpy.nonzero ( regs_mz ) )[1]
        print "regions %d nonzero, %d points" % (nz, len(points))

        if nz != len(points) :
            print "mask failed - %d of %d pts nonzero" % (nz, len(points))

        nz = numpy.shape ( numpy.nonzero ( regs_fz ) )[1]
        print "struct. %d nonzero, %d atoms" % (nz, len(atoms))

        nz_int =  numpy.shape ( (regs_mz * regs_fz).nonzero () )[1]
        nz_uni =  numpy.shape ( (regs_mz + regs_fz).nonzero () )[1]

        sm_score = float(nz_int) / float (nz_uni)

        if 1 or bPrint : print " - intersection %d, union %d" % (nz_int, nz_uni)

        print " * shape match score %.3f" % sm_score

        return sm_score





    def SelRegsShapeScore ( self ) :


        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        smod = self.CurrentSegmentation()
        if smod is None : return

        fmap = self.SimulatedMap()
        if fmap == None:
            return


        regs = smod.selected_regions()
        if len(regs)==0 :
            umsg ( "Please select a region" );
            return

        print "Shape score of %s to %d regions:" % (fmap.name, len(regs)),
        for r in regs :
            print r.rid,
        print ""

        fmap.sms = self.ShapeMatchScore ( fmap.fmol.atoms, dmap, regs )
        ov, fmap.fit_score = map_overlap_and_correlation ( fmap, dmap, True )

        #self.ShapeMaskedCorr ( fmap, dmap, regs )
        #self.OptShapeScore ( fmap, dmap, regs )

        umsg ( "Shape match score: %.3f, cross-correlation: %.3f" % (fmap.sms, fmap.fit_score) )

        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        log_file = path + smod.name + "_fits_sms.txt"
        print "SMS to", log_file

        fpl = open ( log_file, 'a' )
        fpl.write ( "%s %f %f\n" % (fmap.name, fmap.sms, fmap.fit_score) )
        fpl.close ()




    def SelRegsOptimizeShapeScore ( self ) :


        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        smod = self.CurrentSegmentation()
        if smod is None : return

        fmap = self.SimulatedMap()
        if fmap == None:
            return


        regs = smod.selected_regions()
        if len(regs)==0 :
            umsg ( "Please select a region" );
            return

        print "Shape score of %s to %d regions:" % (fmap.name, len(regs)),
        for r in regs :
            print r.rid,
        print ""

        # fmap.sms = self.ShapeMatchScore ( fmap, dmap, regs )
        # self.ShapeMaskedCorr ( fmap, dmap, regs )
        # fmap.sms = self.OptShapeScore ( fmap, dmap, regs )

        umsg ( "Optimized shape match score: %.3f" % fmap.sms )


    def SimulatedMap ( self ):

        if len ( self.struc_map.get() ) == 0 :
            self.GenStrucMap( show = False )
            if len ( self.struc_map.get() ) == 0 :
                return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return
        return fmap


    def MaskedMap ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "Please select a density map"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return


        imap, f_COM, f_bRad = self.MapMaskedMapIndexes ( dmap, fmap, True )


        

    def MapMaskedMapIndexes ( self, ref_map, mask_map, bAddMap=False ) :

        f1 = ref_map.data.ijk_to_xyz_transform
        f2 = Matrix.xform_matrix( ref_map.openState.xform )
        tf = Matrix.multiply_matrices( f2, f1 )

        n1, n2, n3 = ref_map.data.size[0], ref_map.data.size[1], ref_map.data.size[2]
        m_points = VolumeData.grid_indices( (n1,n2,n3), numpy.single )  # i,j,k indices
        _contour.affine_transform_vertices( m_points, tf )

        a_vals = mask_map.interpolated_values ( m_points, chimera.Xform() )
        mm = a_vals.reshape( (n3,n2,n1) )

        thr = mask_map.surface_levels[0]
        mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )
        print "- in masked map, %d above %.3f" % ( numpy.shape(mm.nonzero())[1], thr )

        if bAddMap :
            mmd = VolumeData.Array_Grid_Data (mm, ref_map.data.origin, ref_map.data.step, ref_map.data.cell_angles, ref_map.data.rotation)
            gv = VolumeViewer.volume.add_data_set ( mmd, None )
            gv.name = ref_map.name + "---" + mask_map.name
            #gv.openState.xform = dmap.openState.xform

        nze = numpy.nonzero ( mm )
        print " - index values of %d non-zero points" % len ( nze[0] )

        imap = {}
        for ei, i in enumerate ( nze[0] ) :
            j = nze[1][ei]
            k = nze[2][ei]

            try : mi = imap[i]
            except : mi = {}; imap[i] = mi

            try : mij = mi[j]
            except : mij = {}; mi[j] = mij

            mij[k] = 1


        nzs = numpy.array ( [nze[2], nze[1], nze[0]] )
        points = numpy.cast[numpy.float32] ( numpy.transpose (nzs) )
        _contour.affine_transform_vertices( points, f1 )

        com = numpy.sum(points, axis=0) / len(points)
        C = chimera.Vector ( com[0], com[1], com[2] )
        comv = numpy.ones_like ( points ) * com
        points = points - comv
        bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points), 1 ) ) )

        return imap, C, bRad



    def MapIndexesInMap ( self, ref_map, mask_map ) :

        thr = mask_map.surface_levels[0]
        mm = mask_map.data.matrix()
        mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )

        nze = numpy.nonzero ( mm )

        # copy is needed! _contour.affine_transform requires contiguous array
        points =  numpy.empty ( (len(nze[0]), 3), numpy.float32)
        points[:,0] = nze[2]
        points[:,1] = nze[1]
        points[:,2] = nze[0]

        print "Making map indices for %s in %s" % ( mask_map.name, ref_map.name )
        print " - %d points above %.3f" % ( len(points), thr )

        # transform to index reference frame of ref_map
        f1 = mask_map.data.ijk_to_xyz_transform
        f2 = Matrix.xform_matrix ( mask_map.openState.xform )
        f3 = Matrix.xform_matrix ( ref_map.openState.xform.inverse() )
        f4 = ref_map.data.xyz_to_ijk_transform

        tf = Matrix.multiply_matrices( f2, f1 )
        tf = Matrix.multiply_matrices( f3, tf )
        tf = Matrix.multiply_matrices( f4, tf )
        _contour.affine_transform_vertices ( points, tf )

        imap = set()
        for fi, fj, fk in points :
            for i in [ int(numpy.floor(fi)), int(numpy.ceil(fi)) ] :
                for j in [ int(numpy.floor(fj)), int(numpy.ceil(fj)) ] :
                    for k in [ int(numpy.floor(fk)), int(numpy.ceil(fk)) ] :
                        imap.add((i,j,k))

        return imap




    def OverlappingRegions ( self, dmap, fmap, smod, hide_others = True ) :

        imap = self.MapIndexesInMap ( dmap, fmap ) 
        print 'imap', len(imap)

        try : fmap.COM
        except : fmap.COM = fmap.mol.COM; fmap.bRad = fmap.mol.BoundRad

        p = numpy.array ( [ fmap.COM ], numpy.float32 )
        _contour.affine_transform_vertices ( p, Matrix.xform_matrix( fmap.openState.xform ) )
        _contour.affine_transform_vertices ( p, Matrix.xform_matrix( dmap.openState.xform.inverse() ) )
        f_COM = chimera.Vector ( *p[0] )
        f_bRad = fmap.bRad
        print " center", f_COM, "brad", f_bRad

        jregs = []

        for r in smod.regions :

            if r.placed:
                continue

            ipoints = r.points()
            noverlap = 0
            for i,j,k in ipoints :
                if (i,j,k) in imap:
                    noverlap += 1

            ov = float(noverlap) / len(ipoints)

#            print " region %d, overlap %.3f" % (r.rid, ov)

            if ov > .5 :
                #print " - %d/%d - %d - %d of %d overlap, %.3f" % ( reg_i+1, len(smod.surfacePieces), r.rid, noverlap, r.point_count(), ov )
                #r.display = True
                jregs.append ( r )
            # else : r.display = False

            try :
                if ov > r.max_ov :
                    r.max_ov = ov
                    r.max_ov_cid = fmap.chain_id
                    r.max_ov_bioM = smod.bio_mt_at
            except :
                pass


        oregs = jregs
        
        umsg ( "Found %d regions overlapping" % len(oregs) )
        sel_sps = []

        for sp in smod.surfacePieces :
            try : clr = sp.region.color
            except : continue
            if oregs.count ( sp.region ) > 0 :
                sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
                sp.display = True
                sel_sps.append ( sp )
            else :
                sp.color = ( clr[0], clr[1], clr[2], 1.0 )
                sp.display = not hide_others

        return jregs



    def ZeroDMap_with_FMap ( self ) :
        
        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        print "Taking %s away from %s" % ( fmap.name, dmap.name )
        rname = dmap.name + '_-_' + fmap.name

        mmc = getMod ( rname )

        if mmc == None :
            mmc = dmap.writable_copy ()
            mmc.name = rname
            print " - cloned", dmap.name
        else :
            print " - found", mmc.name

        self.ZeroOverlappingRegion ( mmc, fmap )





    def ZeroOverlappingRegion ( self, ref_map, mask_map ) :

        thr = mask_map.surface_levels[0]
        mm = mask_map.data.matrix()
        mm = numpy.where ( mm > thr, mm, numpy.zeros_like(mm) )

        nze = numpy.nonzero ( mm )
        nzs = numpy.array ( [nze[2], nze[1], nze[0]] )

        # the copy is needed! otherwise the _contour.afine_transform does not work for some reason
        points = numpy.transpose ( nzs ).astype(numpy.float32)

        print " - %s - %d points above %.3f" % ( mask_map.name, len(points), thr )

        # transform to index reference frame of ref_map
        f1 = mask_map.data.ijk_to_xyz_transform
        f2 = Matrix.xform_matrix ( mask_map.openState.xform )
        f3 = Matrix.xform_matrix ( ref_map.openState.xform.inverse() )
        f4 = ref_map.data.xyz_to_ijk_transform

        tf = Matrix.multiply_matrices( f2, f1 )
        tf = Matrix.multiply_matrices( f3, tf )
        tf = Matrix.multiply_matrices( f4, tf )

        _contour.affine_transform_vertices ( points, tf )


        mm = ref_map.data.matrix()
        n1, n2, n3 = ref_map.data.size[0], ref_map.data.size[1], ref_map.data.size[2]

        for fi, fj, fk in points :

            for i in [ int(numpy.floor(fi)), int(numpy.ceil(fi)) ] :
                for j in [ int(numpy.floor(fj)), int(numpy.ceil(fj)) ] :
                    for k in [ int(numpy.floor(fk)), int(numpy.ceil(fk)) ] :
                        if i < 0 or i >= n3 : continue
                        if j < 0 or j >= n2 : continue
                        if k < 0 or k >= n1 : continue
                        mm[k,j,i] = 0.0




    def TakeDMap_with_FMap ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        if len(self.struc_map.get()) == 0 : print "Please select a structure map"; return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : print self.struc_map.get(), "not found"; return

        print "Taking densities from %s with %s" % ( dmap.name, fmap.name )
        rname = fmap.name + '_-_' + dmap.name

        #mmc = fmap.writable_copy ( require_copy = True )
        #mmc.name = rname
        #print " - cloned", fmap.name


        n1, n2, n3 = fmap.data.size[0], fmap.data.size[1], fmap.data.size[2]
        f_points = VolumeData.grid_indices( (n1,n2,n3), numpy.single )  # i,j,k indices
        _contour.affine_transform_vertices( f_points, fmap.data.ijk_to_xyz_transform )

        d_vals = dmap.interpolated_values ( f_points, fmap.openState.xform )
        df_mat = d_vals.reshape( (n3,n2,n1) )

        f_mat = fmap.data.full_matrix()
        f_mask = numpy.where ( f_mat > fmap.surface_levels[0], numpy.ones_like(f_mat), numpy.zeros_like(f_mat) )

        df_mat = df_mat * f_mask

        df_data = VolumeData.Array_Grid_Data ( df_mat, fmap.data.origin, fmap.data.step, fmap.data.cell_angles )
        
        try : df_v = VolumeViewer.volume.add_data_set ( df_data, None )
        except : df_v = VolumeViewer.volume.volume_from_grid_data ( df_data )

        df_v.name = rname
        df_v.openState.xform = fmap.openState.xform


        

    def ShowOverlappingRegions ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return

        if len(self.struc_map.get()) == 0 : umsg ("Please select a structure map"); return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return


        oregs = self.OverlappingRegions ( dmap, fmap, smod )

        return oregs




    def FitMapToGroupsAround ( self, fmap, smod, reg, dmap, bFirst = True ) :

        bestFitScore = -1e99
        bestFitM = None
        bestFitGroup = None
        fmap.fit_score = None

        tvol = self.MapVolume ( fmap )
        bRad = -1.0 # fmap.mol.BoundRad / float(dmap.data.step[0]); # self.MapBoundingRad ( fmap )
        bRad = fmap.mol.BoundRad

        print "\nMaking groups around region %d - target vol %.3f, b-Rad %.3f" % ( reg.rid, tvol, bRad )
        smod.rgroups, maxDepthReached = self.GroupAroundReg ( smod, reg, tvol, bRad )
        smod.rgroups.sort()
        print " - depth reached: %d" % maxDepthReached

        if len(smod.rgroups) == 0 : umsg ( "No groups found" ); return -1e99

        nsearchgrps = min ( len(smod.rgroups), SAF_LS_NGROUPS )
        print "________________________________________________________________________"
        umsg ( "Fitting %s to %d/%d groups..." % ( fmap.name, nsearchgrps, len(smod.rgroups) ) )
        print "________________________________________________________________________"



        for i, dv_regs in enumerate ( smod.rgroups[0:nsearchgrps] ) :

            dv, regs = dv_regs

            umsg ( "Fitting to group %d/%d, dVolume %.4f, %d regions" % (i+1, nsearchgrps, dv, len(regs) ) )
            print " - regions:",
            for r in regs : print r.rid,
            print ""

            for sp in smod.surfacePieces :
                if regs.count ( sp.region ) > 0 :
                    sp.display = True
                    clr = sp.region.color
                    sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

                else : sp.display = False


            points = numpy.concatenate ( [r.map_points() for r in regs], axis=0 )
            self.saFitMapToPoints ( fmap, points, dmap )

            print ""

            if fmap.fit_score > bestFitScore :
                bestFitScore = fmap.fit_score
                bestFitM = fmap.M

        umsg ( "Best cross-correlation: %.4f\n\n" % ( bestFitScore ) )

        fmap.fit_score = bestFitScore
        fmap.M = bestFitM

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA




    def FitMapToRegionsAroundSel ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return


        regs = smod.selected_regions()

        if len(regs) != 1 : umsg ( "Please select 1 region" ); return


        if len(self.struc_map.get()) == 0 : umsg ("Please select a structure map"); return
        fmap = getMod ( self.struc_map.get() )
        if fmap == None : umsg ( "%s is not open" % self.struc_map.get() ); return


        if timing: t0 = clock()
        self.FitMapToGroupsAround ( fmap, smod, regs[0], dmap )
        if timing:
            t1 = clock()
            print "Time: %.1f sec" % (t1-t0,)


        oregs = self.OverlappingRegions ( dmap, fmap, smod, hide_others = False )

        if len(oregs) == 1 :
            oregs[0].placed = True



    def FitMapsToRegionsAroundSel ( self ) :

        print "_______________________________________________________________"

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        smod = self.CurrentSegmentation()
        if smod is None : return


        sregs = smod.selected_regions()        
        if len(sregs) != 1 : print "please selected 1 region"; return

        if timing: t0 = clock()

        for sp in smod.surfacePieces :
            if sp.region == sregs[0].region :
                sp.display = True
                clr = sp.region.color
                sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
            else :
                sp.display = False


        self.FitOpenMapsToGroupsAround ( smod, sregs[0].region, dmap )

        if timing:
            t1 = clock()
            print "Time: %.1f sec" % (t1 - t0)



    def FitOpenMapsToGroupsAround ( self, smod, reg, dmap, bFirst=True ) :

        bestFitScore = -1e99
        bestFitMap = None
        bestFitM = None

        fmaps = []

        # TODO: Don't use "centered" to decide what maps to fit.
        for fmap in OML() :
            try : fmap.mol.centered
            except : continue
            fmap.display = False
            fmap.mol.display = False
            fmaps.append ( fmap )


        for fmap in fmaps :

            print "\n****************************************************"
            print "Fitting: ", fmap.name
            print "****************************************************\n"

            fmap.mol.display = True

            self.FitMapToGroupsAround ( fmap, smod, reg, dmap, bFirst )

            if fmap.fit_score == None :
                print " - no fits for map", fmap.name

            elif fmap.fit_score > bestFitScore :
                bestFitScore = fmap.fit_score
                bestFitMap = fmap
                bestFitM = fmap.M

            fmap.mol.display = False


        if bestFitM == None :
            print "No best fit recorded, perhaps there were not groups to start with"
            return None

        fmap = bestFitMap
        fmap.mol.display = True

        print "\n**********************************************************"
        print "Best fit score was %.4f for %s" % (bestFitScore, fmap.name)
        print "**********************************************************\n"
        fmap.M = bestFitM

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        oregs = self.OverlappingRegions ( dmap, fmap, smod, hide_others = False  )

        if oregs.count ( reg ) == 0 :
            print "Overlapping regions not inclusive"
            return False


        if len(oregs) > 1 :
            jreg = smod.join_regions ( oregs )
            jreg.placed = True
            self.ReportRegionCount(smod)

            jsp = jreg.surface_piece
            clr = jreg.color
            jsp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
            jsp.display = False

        elif len(oregs) == 1 :
            oregs[0].placed = True
            oregs[0].surface_piece.display = False

        else :
            fmap.mol.display = False
            return False

        fmap.mol.display = False

        if self.strucCopyAfterFit.get() :
            self.SaveFit ( fmap )

        return True



    def FitMapsToRGroups ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        smod = self.CurrentSegmentation()
        if smod == None : return

        for sp in smod.surfacePieces :
            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
            sp.display = False


        if timing: t0 = clock()

        while 1 :

            sp = None
            for spi in smod.surfacePieces :

                try :
                    spi.region.placed.display = False

                    clr = spi.region.color
                    spi.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
                    spi.display = True
                    spi.region.placed.display = True
                    spi.display = False

                except :
                    pass

                if spi.region.placed == False :
                    sp = spi
                    break

            if sp == None :
                print "\nAll regions have maps placed"
                break

            clr = sp.region.color
            sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )
            sp.display = True

            if self.FitOpenMapsToGroupsAround ( smod, sp.region, dmap, False ) == False :
                print "___ No map fit for region! ___", sp.region.rid, sp.region.enclosed_volume()
                sp.failed_fit = True

            sp.region.placed = True


        if timing: 
            t1 = clock()
            print "Time: %.1f sec" % (t1-t0)

        for sp in smod.surfacePieces :
            try : sp.failed_fit
            except : continue
            print "Region %d failed fit and still in model" % sp.region.rid







    def MapVolume ( self, fmap ) :

        thr = fmap.surface_levels[0]
        mm = fmap.data.matrix()
        mmab = numpy.where ( mm > thr, numpy.ones_like(mm), numpy.zeros_like(mm) )
        nz = numpy.shape ( numpy.nonzero ( mmab ) )[1]
        vvol = fmap.data.step[0] * fmap.data.step[1] * fmap.data.step[2]
        tvol = vvol * float(nz)
        print "%s - %d above %f, VOLUME %.3f" % (fmap.name, nz, thr, tvol)
        return tvol


    def MapBoundingRad ( self, fmap ) :
        
        fmol = fmap.mol
        points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = False )

        print "%s (%s)\n - COM: %.3f %.3f %.3f, bounding rad %.3f" % ( fmap.name, fmol.name, C[0], C[1], C[2], bRad )

        return bRad




    def FitMapToRGroups ( self ) :

        print "_______________________________________________________________"

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        fmap = self.SimulatedMap()
        if fmap == None : return
        tvol = self.MapVolume ( fmap )
                                         
        smod = self.CurrentSegmentation()
        if smod is None : return

        print "---"

        if timing: t0 = clock()

        bRad = fmap.mol.BoundRad

        smod.rgroups = self.GroupAllRegions ( smod, tvol, bRad )
        smod.rgroups.sort()

        bestFitScore = -1e99
        #bestFitMap = None
        bestFitM = None

        print "Got %d groups..." % (len(smod.rgroups) )

        # return

        dmap_name = os.path.splitext ( dmap.name )[0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        log_f = path + "%s_fits.txt" % (dmap_name)
        print "Log file: %s" % log_f
        ffp = open ( log_f,  'w' ) ; ffp.close()


        nsearchgrps = min ( MAX_NUM_GROUPS, len(smod.rgroups) )

        for i, dv_regs in enumerate ( smod.rgroups [0:nsearchgrps] ) :

            dv, regs = dv_regs

            umsg ( "Fitting to group %d/%d, dVolume %.4f, %d regions" % (i+1, nsearchgrps, dv, len(regs) ) )
            print " - regions:",
            for r in regs : print r.rid,
            print ""

            for sp in smod.surfacePieces :
                if regs.count ( sp.region ) > 0 :
                    sp.display = True
                    clr = sp.region.color
                    sp.color = ( clr[0], clr[1], clr[2], REG_OPACITY )

                else : sp.display = False


            # TODO: points need to be in dmap coordinates.
            points = numpy.concatenate ( [r.map_points()
                                          for r in regs], axis=0 )
            if self.rotaSearch.get () :
                self.saFitMapToPoints_byRot ( fmap, points, dmap )
            else :
                self.saFitMapToPoints ( fmap, points, dmap )

            fmap.sms = 0.0

            fp = open ( log_f, 'a' )
            fp.write ( "%s %f %f %f" % (fmap.mol.name, fmap.fit_score, fmap.sms, dv) )
            T = fmap.M ; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
            T = fmap.MI; fp.write ( " %f %f %f %f %f %f %f %f %f %f %f %f" % (T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3]) )
            for r in regs : fp.write ( " %d" % r.rid )
            fp.write ( "\n" )
            fp.close()


            if fmap.fit_score > bestFitScore :
                bestFitScore = fmap.fit_score
                bestFitM = fmap.M

        umsg ( "Best cross-correlation: %.4f\n\n" % ( bestFitScore ) )

        fmap.fit_score = bestFitScore
        fmap.M = bestFitM

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        if timing: 
            t1 = clock()
            print "Time: %.1f sec" % (t1-t0)

        oregs = self.OverlappingRegions ( dmap, fmap, smod, hide_others = False  )

        self.GetFits ()



    def GetFits ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        dmap_name = os.path.splitext ( dmap.name )[0]
        path = os.path.dirname ( dmap.data.path ) + os.path.sep

        log_f = path + "%s_fits.txt" % (dmap_name)
        print "Log file: %s" % log_f


        class ClusterEntry :
            def __init__ (self, COM, o) :
                self.COM = chimera.Vector ( COM[0], COM[1], COM[2] )
                self.o = o

        class Cluster :
            def __init__ (self) :
                self.entries = []
                self.COM = chimera.Vector (0,0,0)

            def AddEntry ( self, new_e ) :
                self.entries.append  ( [new_e.o.sms,  new_e] )

                # compute the new COM
                self.COM = chimera.Vector (0,0,0)
                for o, e in self.entries :
                    self.COM = self.COM + e.COM
                self.COM = self.COM / float ( len(self.entries) )

        class Fit:
            def __init__ ( self, mname, cor, sms, dv ):
                self.mname = mname
                self.cor = cor
                self.sms = sms
                self.dv = dv

        try : fp = open ( log_f, 'r' )
        except :
            print "Could not open log file:", log_f
            return

        self.clusters = []

        li = 0
        while 1 :
            li = li + 1
            line = fp.readline()
            if len(line) == 0 : break

            n = line.split()

            o = Fit ( n[0], float(n[1]), float(n[2]), float(n[3]) )

            # get M
            mi = 4
            M = numpy.matrix ( [ [ float(n[mi+ 0]), float(n[mi+ 1]), float(n[mi+ 2]), float(n[mi+ 3]) ],
                                 [ float(n[mi+ 4]), float(n[mi+ 5]), float(n[mi+ 6]), float(n[mi+ 7]) ],
                                 [ float(n[mi+ 8]), float(n[mi+ 9]), float(n[mi+10]), float(n[mi+11]) ],
                                 [ 0.0, 0.0, 0.0, 1.0 ]  ] )
            # orthogonalize M
            T = M ; xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
            mt, mr = xf_2_M ( xf )
            o.M = mt * mr

            COM = chimera.Vector ( M[0,3], M[1,3], M[2,3] )
            #print COM

            # get MI
            mi = mi+12
            MI = numpy.matrix ( [ [ float(n[mi+ 0]), float(n[mi+ 1]), float(n[mi+ 2]), float(n[mi+ 3]) ],
                                  [ float(n[mi+ 4]), float(n[mi+ 5]), float(n[mi+ 6]), float(n[mi+ 7]) ],
                                  [ float(n[mi+ 8]), float(n[mi+ 9]), float(n[mi+10]), float(n[mi+11]) ],
                                  [ 0.0, 0.0, 0.0, 1.0 ]  ] )
            # orthogonalize MI
            T = MI ; xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
            mt, mr = xf_2_M ( xf )
            MI = mt * mr

            mi = mi+12

            o.regs = []
            for rmi in range ( mi, len(n) ) :
                ri = int ( n[rmi] )

            # print "%s - corr %.3f, shape score %.3f, regions:" % (o.mname, o.cor, o.sms), o.regs

            cluster = None
            for cl in self.clusters :
                v = cl.COM - COM
                if v.length < 6.0 :
                    cluster = cl

            if cluster == None :
                #print "+++ Creating new cluster: ",
                cluster = Cluster()
                self.clusters.append ( cluster )

            cluster.AddEntry ( ClusterEntry (COM, o) )
            #print "- cluster COM (%f, %f, %f)" % (cluster.COM[0], cluster.COM[1], cluster.COM[2])

            # self.fits.append ( [o.sms, o] )

        fp.close()

        print "____________ %d alignments, %d clusters ____________" % ( li, len(self.clusters) )

        self.fits = []

        for cl in self.clusters :
            cl.entries.sort()
            cl.entries.reverse()
            sms, cle = cl.entries[0]
            self.fits.append ( [cle.o.cor, cle.o] )

        self.fits.sort()
        self.fits.reverse()

        log_f = path + "%s_fits_sorted.txt" % (dmap_name)
        print "Writing fits to: %s" % log_f
        lfp = open ( log_f, "w" )

        for i, sms_o in enumerate ( self.fits ) :
            sms, o = sms_o

            if i < 20 :
                print "%d - %s - correlation %.3f, dv %.3f" % (
                    i+1, o.mname, o.cor, o.dv),
                #for r in o.regs : print r.rid,
                print ""

            lfp.write ( "%d - structure: %s, cross-correlation: %.3f, dVolume: %.3f\n" % (i+1, o.mname, o.cor, o.dv) )

        print ""

        lfp.close ()

        self.fit_at = -1



    def NextFit ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        self.fit_at = self.fit_at + 1
        sms, o = self.fits[self.fit_at]

        print "Fit %d, %s, corr %.3f, sms %.3f, regions" % (self.fit_at, o.mname, o.cor, o.sms),

        smod = self.CurrentSegmentation()
        if smod == None : return

        for r in smod.surfacePieces : r.display = False
        for r in o.regs : r.display = True; print r.rid,

        print ""

        fmol = getMod ( o.mname )
        if fmol == None : print "Fitted mol %s not open" % o.mname; return

        fmap = None
        for m in chimera.openModels.list() :
            # TODO: Don't use "centered" to decide what maps to fit.
            try : m.mol.centered
            except : continue
            if m.mol.name == o.mname :
                fmap = m
                m.display, m.mol.display = True, True
            else :
                m.display, m.mol.display = False, False

        self.struc.set ( fmap.mol.name )
        self.struc_map.set ( fmap.name )

        if fmap == None :
            print "Fitted map corresponding to %s not open" % o.mname; return

        print "Found fitted map", fmap.name                

        fmap.M = o.M
        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * o.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA



    def BestFits0 ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return
        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        print " - path:", path

        smod = self.CurrentSegmentation()
        if smod == None : return

        for r in smod.surfacePieces : r.display = True

        self.GetFits ()
        regs_claimed = {}
        dmap.fitted_mols = []

        for i, sms_o in enumerate ( self.fits ) :

            sms, o = sms_o

            print "%d - %s - corr %.3f, sm_corr %.3f, regions, dv %.3f" % (
                i+1, o.mname, o.cor, o.sms, o.dv),

            bRegionsClaimed = False
            for r in o.regs :
                print r.rid,
                if regs_claimed.has_key ( r.rid ) :
                    bRegionsClaimed = True
                    print "*",
                else :
                    r.display = False
                    regs_claimed[r.rid] = True

            print ""

            if bRegionsClaimed :
                print " - one or more regions claimed"
                break

            nsp = smod.join_regions ( o.regs )
            self.nRegions.set ( len(smod.surfacePieces) )

            fmol = getMod ( o.mname )
            if fmol == None : print " - fitted mol %s not open" % o.mname; return

            fmap = fitMap ( o.mname )
            if fmap == None : print " - fitted map not open"

            fmap.M = o.M
            tXO, tXR = xf_2_M ( dmap.openState.xform )
            T = tXO * tXR * o.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA
            fmap.mol.openState.xform = xfA

            if 0 :
                self.saFitMapToPoints ( fmap, nsp.tpoints, dmap )
                fmap.sms = self.ShapeMatchScore ( fmap.fmol.atoms, dmap, [nsp] )


            if self.strucCopyAfterFit.get() : self.SaveFit ( fmap )



    def PlaceBestFits ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return
        path = os.path.dirname ( dmap.data.path ) + os.path.sep
        print " - path:", path

        map_name = os.path.splitext ( dmap.name )[0]

        smod = self.CurrentSegmentation()
        if smod == None : return

        self.GetFits ()

        dmap.fitted_mols = []

        scores = []

        for i, sms_o in enumerate ( self.fits ) :

            if i >= int ( self.numFits.get() ) : break

            sms, o = sms_o

            print "%d/%d - %s - corr %.3f, sm_corr %.3f, dv %.3f" % (
                i+1, len(self.fits), o.mname, o.cor, o.sms, o.dv)

            fmol = getMod ( o.mname )
            if fmol == None : print " - fitted mol %s not found" % o.mname; return

            fmap = fitMap ( o.mname )
            if fmap == None : print " - fitted map for %s not found" % fmol.name; return

            fmap.M = o.M
            tXO, tXR = xf_2_M ( dmap.openState.xform )
            T = tXO * tXR * o.M
            xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
            fmap.openState.xform = xfA
            fmap.mol.openState.xform = xfA

            #if self.strucCopyAfterFit.get() :
            self.SaveFit ( fmap )
            

            if 0 :
                oregs = self.OverlappingRegions ( dmap, fmap, smod )
                sms = self.ShapeMatchScore ( fmap.mol.atoms, dmap, oregs )
                print " ---- sms %.4f --- " % (sms)
                scores.append ( sms )

            # if joinRegs : smod.join_regions ( regs )


        if 0 :
            log_f = path + map_name + "_fits_acc.txt"
            print "\nAccuracies log file:", log_f

            try : fp = open ( log_f, "a" )
            except : print " - could not open log file"; return
            fp.write ( "%f %f %f " % ( min(scores), max(scores), sum(scores)/float(len(scores))) )
            for sm_score in scores : fp.write ( "%f " % sm_score )
            fp.write ( "\n" )
            fp.close ()
            




    def AlignToSel ( self ) :

        if len(self.struc.get()) == 0 : print "Please select a structure first"; return
        m = getMod ( self.struc.get() )
        if m == None : print self.struc.get(), "not open"; return

        m_cid = None
        to_cid = None
        m_to = None

        sela = chimera.selection.currentContents()[0]
        for sel_at in sela :
            try : cid = sel_at.residue.id.chainId
            except : pass

            if m_cid == None and sel_at.molecule == m :
                m_cid = cid
                print "Found selected atom in chain %s in %s" % (cid, m.name)
            elif m_to == None and sel_at.molecule != m :
                m_to = sel_at.molecule
                to_cid = cid
                print "Found selected atom in chain %s in %s" % (cid, m_to.name)

        if m_cid == None or m_to == None or to_cid == None :
            print "Please select at least two atoms in the corresponding structures to align"
            return
        
        print "Aligning %s chain %s to %s chain %s" % (m.name, m_cid, m_to.name, to_cid)

        m_seq = m.sequences ( asDict=True ) [m_cid]
        to_seq = m_to.sequences ( asDict=True ) [to_cid]

        print " -- %d residues -- %d residues" % ( len(m_seq.residues), len(to_seq.residues) )

        m_ats, to_ats = AlignChains ( m_seq, to_seq )
        xf, rmsd = chimera.match.matchAtoms ( to_ats, m_ats )

        m_points = _multiscale.get_atom_coordinates ( m_ats, transformed = True )
        f_points = _multiscale.get_atom_coordinates ( to_ats, transformed = True )
        vss = numpy.square ( numpy.subtract ( m_points, f_points ) )
        sums = numpy.sum ( numpy.sum ( vss, axis=1 ) )
        armsd = numpy.sqrt ( sums / float ( len(m_points) ) )

        print " - %d aligned - RMSD: %.4f, RMSD as placed: %.4f" % ( len(m_ats), rmsd, armsd )

        mxf = m_to.openState.xform
        mxf.multiply ( xf )
        m.openState.xform = mxf


        fmap = None
        for om in chimera.openModels.list() :
            try : om_mol = om.mol
            except : continue
            if om_mol == m : fmap = om; break

        if fmap == None :
            print " - no map found for struc"
            return

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        print ""
        umsg ( "Fitting %s and %s into %s" % ( fmap.name, m.name, dmap.name ) )

        fxf = m.openState.xform
        dxf = dmap.openState.xform

        mxf = dxf.inverse()
        mxf.multiply ( fxf ) # fxf = dxf * mxf

        tXO, tXR = xf_2_M ( mxf )
        fmap.M = tXO * tXR

        tXO, tXR = xf_2_M ( dmap.openState.xform )
        T = tXO * tXR * fmap.M
        xfA = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3] )
        fmap.openState.xform = xfA
        fmap.mol.openState.xform = xfA

        fmap.M0 = fmap.M



    def FitRMSD ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        print "Map: %s" % dmap.name
        try : print " - %d fitted molecules" % len(dmap.fitted_mols)
        except : print " - no fitted molecules found"; return


        if len(self.struc.get()) == 0 : print "Please select a structure first"; return
        m = getMod ( self.struc.get() )
        if m == None : print self.struc.get(), "not open"; return

        # try : m.ch_colors
        # except : m.ch_colors = RandColorChains ( m )
        # print "%s - %d chains -" % ( m.name, len(m.ch_colors) ), m.ch_colors.keys()

        print " - aligning %s (%d chains) to %d fit mols" % (
            m.name, len(m.sequences()), len(dmap.fitted_mols) )


        mseqs = m.sequences()
        fmol_msi = {}
        m_atoms = []
        f_atoms = []

        for msi, m_seq in enumerate ( mseqs ) :

            min_fmi = None
            min_rmsd = 1e99
            min_m_atoms = None
            min_f_atoms = None

            print "Chain %s - %d residues" % ( m_seq.chain, len(m_seq.residues) )

            for fmi, fm in enumerate ( dmap.fitted_mols ) :

                try : fmol_msi[fmi]; continue
                except : pass

                f_seq = fm.sequences()[0]
                print " - %d/%d %s, %d residues" % (
                    fmi+1, len(dmap.fitted_mols), dmap.fitted_mols[fmi].name, len(f_seq.residues) ),

                m_ats, f_ats = AlignChains ( m_seq, f_seq )
                xf, rmsd = chimera.match.matchAtoms ( m_ats, f_ats )
                m_points = _multiscale.get_atom_coordinates ( m_ats, transformed = True )
                f_points = _multiscale.get_atom_coordinates ( f_ats, transformed = True )
                vss = numpy.square ( numpy.subtract ( m_points, f_points ) )
                sums = numpy.sum ( numpy.sum ( vss, axis=1 ) )
                armsd = numpy.sqrt ( sums / float ( len(m_points) ) )

                print " %d aligned - RMSD: %.4f, RMSD as placed: %.4f" % (
                    len(m_ats), rmsd, armsd )

                if armsd < min_rmsd :
                    min_rmsd = armsd
                    min_fmi = fmi
                    min_m_atoms = m_ats
                    min_f_atoms = f_ats

            if min_fmi != None :
                print " - min is %d - %s" % (min_fmi, dmap.fitted_mols[min_fmi].name)
                fmol_msi[min_fmi] = msi
                m_atoms = m_atoms + min_m_atoms
                f_atoms = f_atoms + min_f_atoms
            else :
                print " - no maps left to align!"

        print ""
        
        # m.chain_fitmol = {}
        for fmi, msi in fmol_msi.iteritems () :
            print "Chain %s -- fit mol %s" % ( mseqs[msi].chain, dmap.fitted_mols[fmi].name )
            # m.chain_fitmol [ mseqs[msi].chain ] = dmap.fitted_mols[fmi]


        print ""

        xf, rmsd = chimera.match.matchAtoms ( m_atoms, f_atoms )
        m_points = _multiscale.get_atom_coordinates ( m_atoms, transformed = True )
        f_points = _multiscale.get_atom_coordinates ( f_atoms, transformed = True )
        vss = numpy.square ( numpy.subtract ( m_points, f_points ) )
        sums = numpy.sum ( numpy.sum ( vss, axis=1 ) )
        armsd = numpy.sqrt ( sums / float ( len(m_points) ) )

        print "Total %d residues aligned - RMSD: %.4f, RMSD as placed: %.4f" % (
            len(m_atoms), rmsd, armsd )




    def FitRMSDNoRef ( self ) :

        dmap = segmentation_map()
        if dmap == None : print "No segmentation map"; return

        print "Map: %s" % dmap.name
        try : print " - %d fitted molecules" % len(dmap.fitted_mols)
        except : print " - no fitted molecules found"; return


        if len(self.struc.get()) == 0 : print "Please select a structure first"; return
        m = getMod ( self.struc.get() )
        if m == None : print self.struc.get(), "not open"; return

        # try : m.ch_colors
        # except : m.ch_colors = RandColorChains ( m )
        # print "%s - %d chains -" % ( m.name, len(m.ch_colors) ), m.ch_colors.keys()

        print " - aligning %s (%d chains) to %d fit mols" % (
            m.name, len(m.sequences()), len(dmap.fitted_mols) )


        mseqs = m.sequences()
        cmap = []
        for msi, m_seq in enumerate ( mseqs ) :

            min_fid = None
            min_rmsd = 1e99
            min_txf = None
            cmap.append ( '?' )

            for fmi, fm in enumerate ( dmap.fitted_mols ) :

                bIncluded = False
                for cfmi in cmap :
                    if cfmi == fmi : bIncluded = True; break
                if bIncluded : continue

                cmap[msi] = fmi
                # print cmap,

                mats,fats = [],[]
                for i in range ( len(cmap) ) :
                    cid = cmap[i]
                    # print cid, fmols[i].name,
                    for ca_posi, cat in ch_cas[cid].iteritems() :
                        mats.append ( cat.coord() )
                        fats.append ( f_ch_cas[i][ca_posi].coord() )

                # print "matching %d to %d ca atoms" % ( len(fats), len(mats) )
                # txf, rmsd = chimera.match.matchAtoms(fats, mats)
                txf, rmsd = chimera.match.matchPositions(fats, mats)
                # print "- rmsd:", rmsd
                if rmsd < min_rmsd :
                    min_rmsd = rmsd
                    min_cid = chains[i2]
                    min_txf = txf

            print "- pos %d is %s, rmsd %f" % (i1, min_cid, min_rmsd),
            cmap[i1] = min_cid
            print cmap
            # return

            for at in m.atoms :
                c = min_txf.apply ( at.coord() )
                at.setCoord ( c )

        print cmap







def fit_segments_dialog ( create=False ) :

  from chimera import dialogs
  return dialogs.find ( Fit_Segments_Dialog.name, create=create )


def show_fit_segments_dialog ():

    from chimera import dialogs
    d = fit_segments_dialog ( create = True )
    # Avoid transient dialog resizing when created and mapped for first time.
    d.toplevel_widget.update_idletasks ()
    d.enter()
    return d



# -----------------------------------------------------------------------------
#
from chimera import dialogs
dialogs.register (Fit_Segments_Dialog.name, Fit_Segments_Dialog,
                  replace = True)


def fitMap(name):
    for m in chimera.openModels.list() :
        if hasattr(m, 'mol') and m.mol.name == name :
            if not hasattr(m.mol, 'centered'):
                centerMol ( m.mol )
            return m
    return None


def molApplyT ( mol, T ) :

    xf = chimera.Xform.xform ( T[0,0], T[0,1], T[0,2], T[0,3], T[1,0], T[1,1], T[1,2], T[1,3], T[2,0], T[2,1], T[2,2], T[2,3], True )
    # print xf

    mol.COM = chimera.Vector (0,0,0)

    for at in mol.atoms :
        c = xf.apply ( at.coord() )
        at.setCoord ( c )
        mol.COM = mol.COM + c.toVector()

    mol.COM = mol.COM / float ( len(mol.atoms) )

#
# Change atom coordinates so the center of mass is at the origin and
# the principal axes are x, y and z.  Make a compensating transformation
# of the molecule coordinate system so the molecule does not move in the
# graphics window.
#
def centerMol ( fmol ):

        umsg ( "Centering %s" % fmol.name )

#        fmol.ch_colors = RandColorChains ( fmol )
#        fmol.color = fmol.atoms[0].color

        # for at in fmol.atoms : at.mass = at.element.mass

        points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = False )
        fmol.COM, fmol.U, fmol.S, fmol.V = prAxes ( points )

        # move COM to origin and align pr. axes with XYZ
        tAO = numpy.matrix ( [
            [ 1, 0, 0, -fmol.COM[0] ],
            [ 0, 1, 0, -fmol.COM[1] ],
            [ 0, 0, 1, -fmol.COM[2] ],
            [ 0, 0, 0,           1 ]  ] )

        tAR = numpy.matrix ( [
            [ fmol.V[0,0], fmol.V[0,1], fmol.V[0,2], 0 ],
            [ fmol.V[1,0], fmol.V[1,1], fmol.V[1,2], 0 ],
            [ fmol.V[2,0], fmol.V[2,1], fmol.V[2,2], 0 ],
            [           0,           0,           0, 1 ]  ] )

        #print tAO
        #print tAR
        #print tAR * tAO

        molApplyT ( fmol, tAO )
        print " - new COM:", fmol.COM

        molApplyT ( fmol, tAR )
        print " - new COM:", fmol.COM

        # Adjust coordinate system so molecule does not appear to move.
        tf = Matrix.invert_matrix((tAR*tAO).tolist()[:3])
        fmol.openState.localXform(Matrix.chimera_xform(tf))

        fmol.mT = numpy.matrix ( [
            [ 1, 0, 0, 0 ],
            [ 0, 1, 0, 0 ],
            [ 0, 0, 1, 0 ],
            [ 0, 0, 0, 1 ]  ] )

        fmol.mR = numpy.matrix ( [
            [ 1, 0, 0, 0 ],
            [ 0, 1, 0, 0 ],
            [ 0, 0, 1, 0 ],
            [ 0, 0, 0, 1 ]  ] )

        fmol.M = fmol.mT * fmol.mR

        #com = numpy.sum(points, axis=0) / len(points)
        #comv = numpy.ones_like ( points ) * com
        #points0 = points - comv

        points = _multiscale.get_atom_coordinates ( fmol.atoms, transformed = False )
        fmol.COM, fmol.U, fmol.S, fmol.V = prAxes ( points )
        # print fmol.U

        ppoints = points * fmol.U
        fmol.BoundRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (ppoints), 1 ) ) )
        fmol.Extents = numpy.asarray ( numpy.max ( numpy.abs ( ppoints ), 0 ) )[0]

        fmol.Extents[0] = fmol.Extents[0] + 5.0
        fmol.Extents[1] = fmol.Extents[1] + 5.0
        fmol.Extents[2] = fmol.Extents[2] + 5.0

        #print " - b-rad %.3f," % (fmol.BoundRad)
        #print " - extents: %.2f %.2f %.2f" % (fmol.Extents[0], fmol.Extents[1], fmol.Extents[2])

        fmol.centered = True

        umsg ( "Centered %s (radius %.2fA, extents %.2fA %.2fA %.2fA)" % (
            fmol.name, fmol.BoundRad, fmol.Extents[0], fmol.Extents[1], fmol.Extents[2] ) )


def xf_2_M (xf) :

    X = ( numpy.matrix (xf.getOpenGLMatrix()) ).reshape([4,4]).transpose()

    tXO = numpy.matrix ( [
        [ 1, 0, 0, X[0,3] ],
        [ 0, 1, 0, X[1,3] ],
        [ 0, 0, 1, X[2,3] ],
        [ 0, 0, 0,      1 ]  ] )

    tXR = numpy.matrix ( [
        [ X[0,0], X[0,1], X[0,2], 0 ],
        [ X[1,0], X[1,1], X[1,2], 0 ],
        [ X[2,0], X[2,1], X[2,2], 0 ],
        [      0,      0,      0, 1 ]  ] )

    return [tXO, tXR]


def am_2_M ( X ) :

    M = numpy.matrix ( [
        [ X[0,0], X[0,1], X[0,2], X[0,3] ],
        [ X[1,0], X[1,1], X[1,2], X[1,3] ],
        [ X[2,0], X[2,1], X[2,2], X[2,3] ],
        [      0,      0,      0,      1 ]  ] )

    return M



def CopyMol ( mol ) :

    nmol = chimera.Molecule()
    nmol.name = mol.name

    aMap = dict()
    clr = ( rand(), rand(), rand() )

    for res in mol.residues :
        nres = nmol.newResidue (res.type, chimera.MolResId(res.id.chainId, res.id.position))
        # print "New res: %s %d" % (nres.id.chainId, nres.id.position)
        for at in res.atoms :
            nat = nmol.newAtom (at.name, chimera.Element(at.element.number))
            aMap[at] = nat
            nres.addAtom( nat )
            nat.setCoord ( at.coord() )
            nat.drawMode = nat.Sphere
            nat.color = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 )
            nat.display = True

        nres.isHelix = res.isHelix
        nres.isHet = res.isHet
        nres.isSheet = res.isSheet
        nres.isStrand = res.isStrand
        nres.isTurn = res.isTurn
        nres.ribbonDisplay = True
        nres.ribbonDrawMode = 2
        nres.ribbonColor = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 );

    for bond in mol.bonds :
        nb = nmol.newBond ( aMap[bond.atoms[0]], aMap[bond.atoms[1]] )
        nb.display = True

    return nmol



def copyMolChain ( nmol, mol, cid, new_cid, xf, clr ) :

    if nmol == None :
        nmol = chimera.Molecule()
        nmol.name = "complex"

    print "Copying chain %s (%s) to %s" % (cid, mol.name, new_cid)
    aMap = dict()
    for res in mol.residues :
        if res.id.chainId == cid :
            nres = nmol.newResidue(res.type,
                    chimera.MolResId(new_cid, res.id.position))
            # print "New res: %s %d" % (nres.id.chainId, nres.id.position)
            for at in res.atoms :
                nat = nmol.newAtom (at.name, chimera.Element(at.element.number))
                aMap[at] = nat
                nres.addAtom( nat )
                if xf : nat.setCoord ( xf.apply( at.coord() )  )
                else : nat.setCoord ( at.coord() )
                nat.drawMode = nat.Sphere
                nat.color = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 )
                nat.display = True

            nres.isHelix = res.isHelix
            nres.isHet = res.isHet
            nres.isSheet = res.isSheet
            nres.isStrand = res.isStrand
            nres.isTurn = res.isTurn
            nres.ribbonDisplay = True
            nres.ribbonDrawMode = 2
            nres.ribbonColor = chimera.MaterialColor( clr[0], clr[1], clr[2], 1.0 );

    for bond in mol.bonds :
        if (bond.atoms[0].residue.id.chainId == cid and
            bond.atoms[1].residue.id.chainId == cid) :
            nb = nmol.newBond ( aMap[bond.atoms[0]], aMap[bond.atoms[1]] )
            nb.display = True

    return nmol



def TransferChainMap ( a, m ) :

    print " - tr %s" % ( a.name ),

    new_v = m.writable_copy()
    new_v.name = a.name + "_tr"
    mm = new_v.data.full_matrix()

    n1, n2, n3 = m.data.size[0], m.data.size[1], m.data.size[2]
    an1, an2, an3 = a.data.size[0], a.data.size[1], a.data.size[2]

    f_i2s = m.data.ijk_to_xyz_transform
    f_t = Matrix.xform_matrix( m.openState.xform )
    tf = Matrix.multiply_matrices( f_t, f_i2s )

    m_points = VolumeData.grid_indices( (n1,n2,n3), numpy.single )  # i,j,k indices
    _contour.affine_transform_vertices( m_points, tf )

    a_vals = a.interpolated_values ( m_points, chimera.Xform() )
    mm[:,:,:] = a_vals.reshape( mm.shape )

    gv = new_v
    gv.region = ( gv.region[0], gv.region[1], [1,1,1] )

    gv.display = True
    #gv.surface_levels = [ dmap.surface_levels[0] ]
    gv.surface_levels = [ a.d_thr ]
    gv.surf_color = ( rand(), rand(), rand(), 1.0 )
    gv.surface_colors = [ gv.surf_color ]
    ro = VolumeViewer.volume.Rendering_Options()
    #ro.surface_smoothing = True
    #ro.smoothing_factor = .25
    #ro.smoothing_iterations = 10
    gv.update_surface ( False, ro )

    for sp in gv.surfacePieces :
        v, t = sp.geometry
        if len(v) == 8 and len(t) == 12 : sp.display = False


    mmN = numpy.where ( mm > a.d_thr, mm, numpy.zeros_like(mm) )
    nze = numpy.nonzero ( mmN )
    print "%d NZ" % len ( nze[0] )

    gv.imap = {}
    for ei, i in enumerate ( nze[0] ) :
        j = nze[1][ei]
        k = nze[2][ei]

        try : mi = gv.imap[i]
        except : mi = {}; gv.imap[i] = mi

        try : mij = mi[j]
        except : mij = {}; mi[j] = mij

        mij[k] = 1

    nzs = numpy.array ( [nze[2], nze[1], nze[0]] )
    points = numpy.transpose ( nzs ).astype(numpy.float32)
    _contour.affine_transform_vertices ( points, m.data.ijk_to_xyz_transform )

    com = numpy.sum(points, axis=0) / len(points)
    gv.COM = chimera.Vector ( com[0], com[1], com[2] )
    comv = numpy.ones_like ( points ) * com
    points_v = points - comv
    gv.bRad = numpy.sqrt ( numpy.max ( numpy.sum ( numpy.square (points_v), 1 ) ) )

    return gv


def RandColorChains ( m ) :

    ct = {}
    for r in m.residues: ct[r.id.chainId] = 1
    clist = ct.keys()
    clist.sort()
    chains_clrs = {}
    cnames = ""

    for ci, cid in enumerate ( clist ) :
        clr = ( rand()*.7, rand()*.7, rand()*.7 )
        #print "- %s: clr(%.2f, %.2f, %.2f)" % (cid, clr[0], clr[1], clr[2])
        chains_clrs[cid] = chimera.MaterialColor ( clr[0], clr[1], clr[2], 1.0 )
        cnames = cnames + cid

    print "%s - color ribbon for %d chains -" % ( m.name, len(cnames) ), cnames

    # color atoms
    for r in m.residues :
        clr = chains_clrs[r.id.chainId]
        r.ribbonDrawMode = 2
        r.ribbonColor = clr
        r.ribbonDisplay = True
        for at in r.atoms :
            at.display = False
            at.color = clr

    return chains_clrs



def BioMatrices ( m ) :

    matrices = {}

    for rm in m.pdbHeaders['REMARK'] :
        s = rm.split()
        if len(s) == 8 and s[2].find('BIOMT') == 0 :
            # print s
            mi = int ( s[3] )
            ri = int ( s[2][5] ) - 1
            try : matrices[mi]
            except : matrices[mi] = numpy.zeros ( (3,4), numpy.float32 )

            matrices[mi][ri][0] = float ( s[4] )
            matrices[mi][ri][1] = float ( s[5] )
            matrices[mi][ri][2] = float ( s[6] )
            matrices[mi][ri][3] = float ( s[7] )

    return matrices                                       



def map_overlap_and_correlation (map1, map2, above_threshold):

    olap, cor = FitMap.map_overlap_and_correlation ( v1, v2, above_threshold )[:2]
    return olap, cor



def overlap_and_correlation ( v1, v2 ):

    olap, cor = FitMap.overlap_and_correlation ( v1, v2 )[:2]
    return olap, cor



def reportFitRegions(map_name, regs):

    r = ', '.join(str(reg.rid) for reg in regs[:5])
    if len(regs) > 5:
        r += '...'
    umsg ( "Fitting %s to %d regions (%s)" % ( map_name, len(regs), r ) )


def RMSD ( self, ress, fmol ) :

    N, D, posr = 0.0, 0.0, {}
    for r in ress : posr[r.id.position] = r
    for r in fmol.residues :
        v = r.atomsMap["CA"][0].xformCoord() - posr[r.id.position].atomsMap["CA"][0].xformCoord()
        N, D = N + 1, D + (v.length * v.length)

    return numpy.sqrt ( D / N )



def AlignChains ( ref, match ) :

    import MatchMaker
    import MatchMaker.prefs
    import chimera.misc

    gapOpen = MatchMaker.prefs.defaults[MatchMaker.prefs.GAP_OPEN]
    gapExtend = MatchMaker.prefs.defaults[MatchMaker.prefs.GAP_EXTEND]
    ksdsspCache = set()
    ssFraction = MatchMaker.prefs.defaults[MatchMaker.prefs.SS_MIXTURE]
    ssMatrix = MatchMaker.prefs.defaults[MatchMaker.prefs.SS_SCORES]
    alignKw = { 'ssFraction': ssFraction, 'ssMatrix': ssMatrix, 'computeSS': False }

    # print "Aligning %d res to %d res" % ( len(ref), len(match) )

    score, s1, s2 = MatchMaker.align( ref, match, 'BLOSUM-62', "nw", gapOpen, gapExtend, ksdsspCache, **alignKw)

    refAtoms, matchAtoms = [], []

    for i in range( len(s1) ) :
        if s1[i] == "." or s2[i] == ".": continue
        refRes = s1.residues[s1.gapped2ungapped(i)]
        matchRes = s2.residues[s2.gapped2ungapped(i)]
        refAtom = chimera.misc.principalAtom(refRes)
        if not refAtom: continue
        matchAtom = chimera.misc.principalAtom(matchRes)
        if not matchAtom: continue
        if refAtom.name != matchAtom.name:
            # nucleic P-only trace vs. full nucleic
            if refAtom.name != "P":
                try : refAtom = refAtom.residue.atomsMap["P"][0]
                except KeyError: continue
            else:
                try : matchAtom = matchAtom.residue.atomsMap["P"][0]
                except KeyError: continue
        refAtoms.append(refAtom)
        matchAtoms.append(matchAtom)

    # print " - aligned %d atoms" % len (refAtoms)

    if len ( refAtoms ) < 3 :
        print "too few atoms to perform 3D alignment"
        return None

    # return chimera.match.matchAtoms ( refAtoms, matchAtoms )

    return refAtoms, matchAtoms
