# ------------------------------------------------------------------------------
# segment copygroups #2 #3
#

# -----------------------------------------------------------------------------
#
def segment_command(cmdname, args):

    from Commands import parse_arguments, CommandError, string_arg
    if not args.startswith('copygroups'):
      raise CommandError, 'First argument must be "copygroups"'

    req_args = (('operation', string_arg),
                ('from_seg', segmentation_arg),
                ('to_seg', segmentation_arg))
    opt_args = ()
    kw_args = ()
    kw = parse_arguments(cmdname, args[1:], req_args, opt_args, kw_args)
    copy_groups(**kw)


# -----------------------------------------------------------------------------
#
def copy_groups(operation, from_seg, to_seg):

  # Find transform from to_seg mask indices to from_seg mask indices.
  from Matrix import multiply_matrices, invert_matrix, xform_matrix
  tf = multiply_matrices(invert_matrix(from_seg.point_transform()),
                         invert_matrix(xform_matrix(from_seg.openState.xform)),
                         xform_matrix(to_seg.openState.xform),
                         to_seg.point_transform())

  # Find from_seg regions containing to_seg region maximum points.
  fmask = from_seg.mask
  rlist = list(to_seg.regions)
  ijk = [r.max_point for r in rlist]
  from _interpolate import interpolate_volume_data
  rnums, outside = interpolate_volume_data(ijk, tf, fmask, 'nearest')

  # Find to_seg regions that have max_point in same from_seg region.
  fr2tr = {}
  id2r = from_seg.id_to_region
  for tr, frnum in zip(rlist, rnums):
    if frnum > 0:
      fr = id2r[frnum].top_parent()
      if fr in fr2tr:
        fr2tr[fr].append(tr)
      else:
        fr2tr[fr] = [tr]
  groups = [g for g in fr2tr.values() if len(g) > 1]

  # Form groups.
  for g in groups:
    r = to_seg.join_regions(g)
    for gr in g:
        gr.set_color(r.color)

  print 'Made %d groups for %s matching %s' % (len(groups), to_seg.name, from_seg.name)
  

# -----------------------------------------------------------------------------
#
def segmentation_arg(s):

    from chimera.specifier import evalSpec
    sel = evalSpec(s)
    mlist = sel.models()
    from regions import Segmentation
    slist = [s for s in mlist if isinstance(s, Segmentation)]
    if len(slist) == 0:
        raise CommandError, 'No segmentation specified'
    elif len(slist) > 1:
        raise CommandError, 'Multiple segmentations specified'
    return slist[0]
